from .. import core, maps
import numpy as np

__all__ = [
    "CompactMaps",
    "RemoveWeights",
    "ApplyWeights",
    "SetPolConv",
    "FlattenPol",
    "MakeMapsPolarized",
    "MakeMapsUnpolarized",
    "ValidateMaps",
    "ExtractMaps",
    "InjectMapStub",
    "InjectMaps",
    "ReplicateMaps",
    "CoaddMaps",
    "RebinMaps",
    "ReprojectMaps",
    "coadd_map_files",
]


@core.indexmod
def CompactMaps(frame, zero_nans=False):
    """
    Compact all maps in a frame to their default sparse representation.
    Optionally remove NaN values as well.  Removing NaN values will reduce
    memory use, but will remove the distinction in unweighted (or
    weight-removed) maps between unobserved regions and regions with zero
    temperature.
    """
    for s in ["T", "Q", "U", "Wunpol", "Wpol", "H"]:
        if s in frame:
            m = frame.pop(s)
            m.compact(zero_nans=zero_nans)
            frame[s] = m
    return frame


@core.indexmod
def RemoveWeights(frame, zero_nans=False):
    """
    Remove weights from input maps.  If zero_nans is `True`, empty pixels are
    skipped and pixels with zero weight are set to 0 instead of NaN.  Operation
    is performed in place to minimize memory use.
    """
    if "Wpol" not in frame and "Wunpol" not in frame:
        return

    if not frame["T"].weighted:
        return frame
    ValidateMaps(frame)

    tmap = frame.pop("T")

    if "Wpol" in frame:
        wmap = frame["Wpol"]
        qmap = frame.pop("Q")
        umap = frame.pop("U")
        maps.remove_weights(tmap, qmap, umap, wmap, zero_nans=zero_nans)
    else:
        wmap = frame["Wunpol"]
        maps.remove_weights_t(tmap, wmap, zero_nans=zero_nans)

    frame["T"] = tmap
    if "Wpol" in frame:
        frame["Q"] = qmap
        frame["U"] = umap

    return frame


@core.indexmod
def ApplyWeights(frame):
    """
    Apply weights to the input maps.  The operation is performed in place to
    minimize memory use.
    """
    if "Wpol" not in frame and "Wunpol" not in frame:
        return

    if frame["T"].weighted:
        return frame
    ValidateMaps(frame)

    tmap = frame.pop("T")

    if "Wpol" in frame:
        wmap = frame["Wpol"]
        qmap = frame.pop("Q")
        umap = frame.pop("U")
        maps.apply_weights(tmap, qmap, umap, wmap)
    else:
        wmap = frame["Wunpol"]
        maps.apply_weights_t(tmap, wmap)

    frame["T"] = tmap
    if "Wpol" in frame:
        frame["Q"] = qmap
        frame["U"] = umap

    return frame


@core.indexmod
def SetPolConv(frame, pol_conv=maps.MapPolConv.IAU):
    """
    Set or change the polarization convention of the input polarized map frame.
    If switching between IAU and COSMO conventions, flip the sign of the U map
    and the TU and QU weights.  Otherwise, just set the polarization convention
    for all maps and weights.
    """
    if not ("Q" in frame and "U" in frame):
        # only polarized frames
        return frame

    if pol_conv == maps.MapPolConv.none or pol_conv is None:
        raise ValueError("Polarized maps must have pol_conv set to IAU or COSMO")

    tmap = frame.pop("T")
    tmap.pol_conv = pol_conv

    qmap = frame.pop("Q")
    qmap.pol_conv = pol_conv

    umap = frame.pop("U")
    flip = umap.polarized and umap.pol_conv != pol_conv
    umap.pol_conv = pol_conv

    wmap = None
    if "Wpol" in frame:
        wmap = frame.pop("Wpol")
        for k in wmap.keys():
            wmap[k].pol_conv = pol_conv

    # flip sign if switching conventions
    if flip:
        umap *= -1.0
        if wmap is not None:
            wmap.TU *= -1.0
            wmap.QU *= -1.0

    frame["T"] = tmap
    frame["Q"] = qmap
    frame["U"] = umap
    if wmap is not None:
        frame["Wpol"] = wmap

    return frame


@core.indexmod
def FlattenPol(frame, invert=False):
    """
    For maps defined on the sphere the direction of the polarization angle is
    is defined relative to the direction of North.  When making maps we follow
    this definition.

    For any flat sky estimators, the polarization angle is defined relative to
    the vertical axis.  For some map projections the direction of north is not
    the same as the vertical axis.  This function applies a rotation to the Q
    and U values to switch the curved sky Q/U definition to the flat sky Q/U
    definition.

    If for whatever reason you want to reverse the process set the invert
    argument to True.
    """

    if "Q" not in frame or "U" not in frame:
        return

    if any(not isinstance(frame[k], maps.FlatSkyMap) for k in "QU"):
        return

    ValidateMaps(frame, ignore_missing_weights=True)
    tmap, qmap, umap = frame.pop("T"), frame.pop("Q"), frame.pop("U")

    if "Wpol" in frame:
        wmap = frame.pop("Wpol")
        maps.flatten_pol(qmap, umap, wmap, invert=invert)
        frame["Wpol"] = wmap
    else:
        maps.flatten_pol(qmap, umap, invert=invert)

    tmap.flat_pol = not invert
    
    frame["T"] = tmap
    frame["Q"] = qmap
    frame["U"] = umap

    return frame


@core.indexmod
def MakeMapsPolarized(frame, pol_conv=maps.MapPolConv.IAU):
    """
    Converts individual unpolarized maps to polarized versions of the same map,
    with the given polarization convention

    This module is only a shim that creates null Q and U maps and populates
    a properly invertible Wpol array from the TT Wunpol weights.
    """
    if "Wunpol" not in frame:
        return

    wgt = frame["Wunpol"].TT
    wgt.pol_conv = pol_conv
    del frame["Wunpol"]

    qmap = frame["T"].clone(False)
    qmap.pol_type = maps.MapPolType.Q
    qmap.pol_conv = pol_conv
    frame["Q"] = qmap
    umap = frame["T"].clone(False)
    umap.pol_type = maps.MapPolType.U
    umap.pol_conv = pol_conv
    frame["U"] = umap
    mask = wgt.to_mask().to_map()

    wgt_out = maps.G3SkyMapWeights()
    wgt_out.TT = wgt
    wgt_out.TQ = wgt.clone(False)
    wgt_out.TU = wgt.clone(False)
    wgt_out.QQ = mask
    wgt_out.QU = wgt.clone(False)
    wgt_out.UU = mask.clone(True)

    for k in wgt_out.keys():
        wgt_out[k].pol_type = getattr(maps.MapPolType, k)

    frame["Wpol"] = wgt_out

    return frame


@core.indexmod
def MakeMapsUnpolarized(frame):
    """
    Converts individual polarized maps to temperature-only versions of the same map.
    """
    if "Wpol" not in frame:
        return

    tmap = frame.pop("T")
    tmap.pol_conv = maps.MapPolConv.none
    frame["T"] = tmap

    wgt = frame.pop("Wpol").TT
    wgt.pol_conv = maps.MapPolConv.none
    wgt_out = maps.G3SkyMapWeights()
    wgt_out.TT = wgt
    frame["Wunpol"] = wgt_out

    del frame["Q"]
    del frame["U"]

    return frame


@core.indexmod
def ValidateMaps(frame, ignore_missing_weights=False):
    """
    Validate that the input map frame has all the necessary keys.

    If ignore_missing_weights is False (default), a warning is issued when the
    frame contains weighted Stokes maps without a weights map.  Set this option
    to True when feeding single bolometer map frames with common weights through
    a pipeline.
    """

    if isinstance(frame, core.G3Frame) and frame.type != core.G3FrameType.Map:
        return

    map_id = frame.get("Id", None)

    if "T" not in frame:
        if "H" in frame:
            return
        core.log_fatal("Map frame %s: Missing T map" % map_id, unit="ValidateMaps")
    if ("Q" in frame and not "U" in frame) or ("U" in frame and not "Q" in frame):
        core.log_fatal("Map frame %s: Missing Q or U map" % map_id, unit="ValidateMaps")
    if "Wpol" in frame and "Wunpol" in frame:
        core.log_fatal(
            "Map frame %s: Found both polarized and unpolarized weights" % map_id,
            unit="ValidateMaps",
        )

    check_weights = False

    stub = frame["T"].clone(False)
    for k in ["T", "Q", "U", "Wpol", "Wunpol", "H"]:
        if k not in frame:
            continue
        if not frame[k].compatible(stub):
            core.log_fatal(
                "Map frame %s: Map %s not compatible with T map" % (map_id, k),
                unit="ValidateMaps",
            )

        if k in ["Wpol", "Wunpol"]:
            if frame[k].TT.pol_type == maps.MapPolType.TT:
                continue
            # set weights polarization properties
            w = frame.pop(k)
            for wk in w.keys():
                w[wk].pol_type = getattr(maps.MapPolType, wk)
                if k == "Wpol":
                    w[wk].pol_conv = frame["U"].pol_conv
            frame[k] = w

        if k in "TQU":
            if k == "U":
                if isinstance(frame[k], maps.FlatSkyMap) and (
                    frame[k].flat_pol != frame["Q"].flat_pol
                ):
                    core.log_fatal(
                        "Map frame %s: Q and U maps have different flat_pol" % map_id,
                        unit="ValidateMaps",
                    )
            if k in "QU":
                if not frame[k].polarized:
                    core.log_warn(
                        "Map frame %s: %s map polarization convention not set" % (map_id, k),
                        unit="ValidateMaps",
                    )
            if frame[k].weighted and not ignore_missing_weights:
                if "Wpol" not in frame and "Wunpol" not in frame:
                    if not check_weights:
                        core.log_warn(
                            "Map frame %s: Missing weights" % map_id,
                            unit="ValidateMaps",
                        )
                        check_weights = True
                else:
                    if k == "T" and "Q" not in frame and "Wunpol" not in frame:
                        core.log_warn(
                            "Map frame %s: Missing unpolarized weights" % map_id,
                            unit="ValidateMaps",
                        )
                    elif k == "Q" and "Wpol" not in frame:
                        core.log_warn(
                            "Map frame %s: Missing polarized weights" % map_id,
                            unit="ValidateMaps",
                        )

        elif k == "H":
            continue

        elif frame[k].polarized:
            if "Q" not in frame or "U" not in frame:
                core.log_fatal(
                    "Map frame %s: Found unpolarized maps with polarized weights" % map_id,
                    unit="ValidateMaps",
                )
            if frame[k].pol_conv != frame["U"].pol_conv:
                core.log_fatal(
                    "Map frame %s: %s and U maps have different pol_conv" % (map_id, k),
                    unit="ValidateMaps",
                )
            if isinstance(frame[k].QQ, maps.FlatSkyMap):
                if frame[k].flat_pol != frame["Q"].flat_pol:
                    core.log_fatal(
                        "Map frame %s: %s and U maps have different flat_pol" % (map_id, k),
                        unit="ValidateMaps",
                    )

        elif "Q" in frame or "U" in frame:
            core.log_fatal(
                "Map frame %s: Found polarized maps with unpolarized weights" % map_id,
                unit="ValidateMaps",
            )


@core.indexmod
class ExtractMaps(object):
    """
    Cache maps that come through the pipeline. Initialize an instance of this
    module before adding to a pipeline.  Any map frames that pass through the pipe
    are stored in the .maps attribute of the object after the pipeline is run.

    Arguments
    ---------
    map_id : string
        If supplied, select only map frames that match this ID.
    copy : bool
        If True, make a copy of the map on extraction.
    ignore_missing_weights : bool
        If False (default), a warning is issued when the frame contains weighted
        Stokes maps without a weights map.  Set this option to True when feeding
        single bolometer map frames with common weights through a pipeline.
    """

    def __init__(self, map_id=None, copy=False, ignore_missing_weights=False):
        self.map_id = map_id
        self.copy_ = copy
        self.ignore_missing_weights = ignore_missing_weights

    @property
    def maps(self):
        """
        Dictionary of map frames, keyed by Id.  Map objects are extracted from each
        frame in the pipeline, and stored as simple dictionaries.  Multiple frames
        with the same Id will result in a list of dictionaries for that Id.
        """
        if not hasattr(self, "_maps"):
            self._maps = {}
        return self._maps

    def __call__(self, frame):
        if frame.type != core.G3FrameType.Map:
            return
        if self.map_id and frame["Id"] != self.map_id:
            return

        ValidateMaps(frame, ignore_missing_weights=self.ignore_missing_weights)

        mid = frame["Id"]
        mdict = {}
        for k in ["T", "Q", "U", "Wpol", "Wunpol", "H"]:
            if k not in frame:
                continue
            mdict[k] = frame[k] if not self.copy_ else frame[k].copy()

        if mid not in self.maps:
            self.maps[mid] = mdict
            return

        if isinstance(self.maps[mid], dict):
            self.maps[mid] = [self.maps[mid], mdict]
            return

        self.maps[mid].append(mdict)


@core.indexmod
class InjectMapStub(object):
    """
    Inject a new map frame from a map stub.

    Arguments
    ---------
    map_id : string
        Id to assign to the new map frame
    map_stub : G3SkyMap instance
        Map stub from which to clone the Stokes maps and weights.  If the
        `weighted` attribute of the stub is True, the output frame will include
        weights.  If the `pol_conv` attribute of the stub is not None, the
        output frame will include Q and U maps (and polarized weights).
    """

    def __init__(self, map_id, map_stub):
        self.map_frame = core.G3Frame(core.G3FrameType.Map)
        self.map_frame["Id"] = map_id

        map_stub = map_stub.clone(False)
        weighted = map_stub.weighted
        polarized = map_stub.polarized

        T = map_stub.clone(False)
        T.pol_type = maps.MapPolType.T
        self.map_frame["T"] = T
        if polarized:
            Q = map_stub.clone(False)
            Q.pol_type = maps.MapPolType.Q
            self.map_frame["Q"] = Q
            U = map_stub.clone(False)
            U.pol_type = maps.MapPolType.U
            self.map_frame["U"] = U
        if weighted:
            W = maps.G3SkyMapWeights(map_stub)
            self.map_frame["Wpol" if polarized else "Wunpol"] = W

    def __call__(self, frame):
        if self.map_frame is None:
            return

        map_frame = self.map_frame
        self.map_frame = None
        return [map_frame, frame]


@core.indexmod
class InjectMaps(object):
    """
    Inject a set of maps into a new map frame.

    Arguments
    ---------
    map_id : string
        Id to assign to the new map frame
    maps_in : list or dict
        Maps to add to the frame.  If a list, contains Stokes maps with valid
        pol_type and weights.  If a dict, contains Stokes and weights maps keyed
        by the standard map frame names.
    ignore_missing_weights : bool
        Skip warning about missing weights.  Useful for masks.
    """

    def __init__(self, map_id, maps_in, ignore_missing_weights=False):
        self.map_frame = core.G3Frame(core.G3FrameType.Map)
        self.map_frame["Id"] = map_id

        if isinstance(maps_in, list):
            for m in maps_in:
                if isinstance(m, maps.G3SkyMap):
                    k = str(m.pol_type)
                    if k not in "TQU":
                        raise ValueError("Input map has invalid pol_type %s" % k)
                    self.map_frame[k] = m
                elif isinstance(m, maps.G3SkyMapWeights):
                    self.map_frame["Wpol" if m.polarized else "Wunpol"] = m
                else:
                    raise TypeError("maps_in must be G3SkyMap or G3SkyMapWeights")

        elif isinstance(maps_in, dict):
            for k, m in maps_in.items():
                if k not in ["T", "Q", "U", "Wpol", "Wunpol", "H"]:
                    continue
                self.map_frame[k] = m

        else:
            raise TypeError("maps_in must be a list or dict")

        ValidateMaps(self.map_frame, ignore_missing_weights=ignore_missing_weights)

    def __call__(self, frame):
        if self.map_frame is None:
            return

        map_frame = self.map_frame
        self.map_frame = None
        return [map_frame, frame]


@core.indexmod
def ReplicateMaps(frame, input_map_id, output_map_ids, copy_weights=False):
    """
    Clone the input map frame with Id input_map_id into new stub frames, one for
    each Id listed in output_map_ids.

    Arguments
    ---------
    input_map_id : string
        ID of the map frame to replicate.  The input frame is discarded after
        replication.
    output_map_ids : list of strings
        List of IDs to assign to replicated map frames.
    copy_weights : bool
        If False, only the first output frame in the list includes a weights key
        (Wpol or Wunpol).  If True, all output frames include a weights key.
    """

    if frame.type != core.G3FrameType.Map:
        return

    if frame["Id"] != input_map_id:
        return

    ValidateMaps(frame)

    frames = []

    first = True
    for oid in output_map_ids:
        fr = core.G3Frame(core.G3FrameType.Map)
        fr["Id"] = oid
        if copy_weights or first:
            map_keys = ["T", "Q", "U", "Wpol", "Wunpol", "H"]
            first = False
        else:
            map_keys = ["T", "Q", "U"]

        for k in map_keys:
            if k not in frame:
                continue
            fr[k] = frame[k].clone(False)

        frames.append(fr)

    return frames


@core.indexmod
class CoaddMaps(object):
    """
    Coadd maps and weights, optionally collating by map Id.  This class can be
    used as an argument to ``pipe.Add()`` as a standard pipeline module, or
    instantiated as a standalone instance.  In the latter case, the object is
    treated as a callable for input map frames, and the ``coadd_frame`` or
    ``coadd_frames`` attribute contains the running coadd(s).

    The output coadd frames contain two additional keys: ``'InputMapIds'`` and
    ``'InputFiles'``, which are both lists of unique map Id's and filenames that
    are associated with the frames that contribute to each coadd.  When one
    coadd is added to another, these keys are updated recursively, so that the
    resulting coadd includes the Id's and filenames that contributed to both
    itself and the other coadd.  The list of filenames can be populated by
    combining this module with a G3Reader whose ``track_filename`` option is set
    to True; however, this feature is fragile and may not work as expected with
    complex pipelines.
    """

    def __init__(
        self,
        map_ids=None,
        output_map_id="Coadd",
        collate=False,
        weighted=True,
        ignore_missing_weights=False,
        drop_input_frames=False,
        record_obs_id=False,
        keep_outputs=False,
    ):
        """
        Arguments
        ---------
        map_ids : list of str
            List of map Id's to include in the coadd(s).  If None, any maps in the
            pipeline are included.  Otherwise, the output of the ``get_map_ids``
            method is compared with this list, and the input frame is discarded if
            no match is found.
        output_map_id : str
            Id to assign to the output frame.  If ``collate`` is True, this argument
            is required and treated as a prefix to which each input map Id is
            appended.
        collate : bool
            If True, coadd unique map Id's into separate output map frames.
        weighted : bool
            If True (default), ensure that maps have had weights applied before
            coadding.  Otherwise, coadd maps without checking the weights.
        ignore_missing_weights : bool
            If False (default), a warning is issued when the frame contains weighted
            Stokes maps without a weights map.  Set this option to True when feeding
            single bolometer map frames with common weights through a pipeline.
        drop_input_frames : bool
            If True, drop input map frames from the pipeline that are included in
            any coadds.
        record_obs_id : bool
            If True, include source name and observation ID info in the output coadd
            frame ``InputMapIds`` key, along with the map ID for each input frame.
            If False, only the map frame ID is included.
        keep_outputs : bool
            If True, preserve the coadd frame module attributes after parsing an
            EndProcessing frame.  Otherwise, coadd frames are deleted to avoid
            memory leaks in pipelines, under the assumption that ownership of the
            frame(s) is transfered to downstream modules, e.g. G3Writer.
        """
        if isinstance(map_ids, str):
            map_ids = [map_ids]
        self.map_ids = map_ids
        self.collate = collate
        if self.collate:
            self.output_map_id = output_map_id
        else:
            self.coadd_frame["Id"] = output_map_id
        self.weighted = weighted
        self.ignore_missing_weights = ignore_missing_weights
        self.drop_input_frames = drop_input_frames
        self.obs_id = None if record_obs_id else False
        self.keep_outputs = keep_outputs

    @property
    def coadd_frame(self):
        """
        Output coadd map frame, also injected into the pipeline on
        EndProcessing.  This attribute is only populated if the ``collate``
        option is set to False.
        """
        if self.collate:
            raise ValueError("Collation is enabled, use .coadd_frames instead.")
        if not hasattr(self, "_coadd_frame"):
            self._coadd_frame = core.G3Frame(core.G3FrameType.Map)
        return self._coadd_frame

    @property
    def coadd_frames(self):
        """
        Dictionary of output coadd map frames, keyed by input map Id.  Each
        frame is also injected into the pipeline on EndProcessing.  This
        attribute is only populated if the ``collate`` option is set to True.
        """
        if not self.collate:
            raise ValueError("Collation is disabled, use .coadd_frame instead.")
        if not hasattr(self, "_coadd_frames"):
            self._coadd_frames = dict()
        return self._coadd_frames

    def reset(self):
        """
        Clear internal frame cache.
        """
        if self.collate:
            self.coadd_frames.clear()
        else:
            delattr(self, "_coadd_frame")

    def get_map_id(self, frame):
        """
        Return Id associated with the input frame

        By default, this returns the "Id" entry in the frame, None if not found.
        Subclass the CoaddMaps structure to override the default behavior.

        Arguments
        ---------
        frame : core.G3Frame instance of type G3FrameType.Map
            Candidate map frame to include in a coadd.

        Returns
        -------
        map_id : str, None or False
            A string if the frame is to be included in a coadd, None if the
            frame is to be passed on to downstream pipeline modules, or False if
            the frame is to be dropped from the pipeline altogether.
        """
        return frame.get("Id", None)

    def __call__(self, frame):

        if isinstance(frame, core.G3Frame) and frame.type == core.G3FrameType.EndProcessing:
            if self.collate:
                out = list(self.coadd_frames.values()) + [frame]
            else:
                out = [self.coadd_frame, frame]
            if not self.keep_outputs:
                self.reset()
            return out

        if self.obs_id is not False and "SourceName" in frame:
            self.obs_id = "{}/{}".format(
                frame["SourceName"], frame["ObservationID"]
            )

        if isinstance(frame, core.G3Frame) and frame.type != core.G3FrameType.Map:
            return

        map_id = self.get_map_id(frame)
        if map_id is None or map_id is False:
            return map_id

        if self.map_ids is not None and map_id not in self.map_ids:
            return

        ValidateMaps(frame, ignore_missing_weights=self.ignore_missing_weights)
        if self.weighted:
            ApplyWeights(frame)

        if self.collate:
            if map_id not in self.coadd_frames:
                fr = core.G3Frame(core.G3FrameType.Map)
                fr["Id"] = self.output_map_id + map_id
                self.coadd_frames[map_id] = fr
            cfr = self.coadd_frames[map_id]
        else:
            cfr = self.coadd_frame

        if "InputMapIds" in cfr:
            map_ids = list(cfr.pop("InputMapIds"))
        else:
            map_ids = []
        if "InputMapIds" in frame:
            # allow for recursive coadds
            map_ids += [i for i in frame["InputMapIds"] if i not in map_ids]
        elif frame.get("Id", None):
            mid = frame["Id"]
            if self.obs_id:
                mid = "{}/{}".format(self.obs_id, mid)
            if mid not in map_ids:
                map_ids += [mid]
        if len(map_ids):
            cfr["InputMapIds"] = core.G3VectorString(map_ids)

        if "InputFiles" in cfr:
            input_files = list(cfr.pop("InputFiles"))
        else:
            input_files = []
        if "InputFiles" in frame:
            # allow for recursive coadds
            input_files += [f for f in frame["InputFiles"] if f not in input_files]
        elif getattr(frame, "_filename", None):
            if frame._filename not in input_files:
                input_files += [frame._filename]
        if len(input_files):
            cfr["InputFiles"] = core.G3VectorString(input_files)

        for key in ["T", "Q", "U", "Wpol", "Wunpol", "H"]:
            if key not in frame:
                continue
            if key not in cfr:
                cfr[key] = frame[key].clone(False)
            m = cfr.pop(key)
            m += frame[key]
            cfr[key] = m

        if self.drop_input_frames:
            return False


@core.usefulfunc
def coadd_map_files(
    input_files,
    output_file=None,
    coadder=None,
    map_ids=None,
    output_map_id="Coadd",
    collate=False,
    weighted=True,
    record_obs_id=False,
):
    """
    Coadd map files, optionally collating map Id's into separate frames.

    Arguments
    ---------
    input_files : list of str
        List of input files to feed through the pipeline.
    output_file : str
        Output G3 filename.  If not supplied, the output frames are
        returned without saving to disk.
    coadder : CoaddMaps instance
        If set, use this instantiated module in the coadding pipeline.
        In this case, all other keyword arguments below are ignored.
    map_ids : list of str
        A list of map Id's to include in the coadd(s).
    output_map_id : str
        Id to use for the output map frame.  If ``collate`` is True,
        this is the prefix applied to each output frame, with the
        input map Id appended to it.
    collate : bool
        If True, coadd individual map Id's into separate map frames.
        Otherwise, all map frames are coadded into one output frame.
    weighted : bool
        If True, ensure that weights have been applied before coadding.
        Otherwise, the input maps are coadded as they are.
    record_obs_id : bool
        If True, include source name and observation ID info in the output coadd
        frame ``InputMapIds`` key, along with the map ID for each input frame.
        If False, only the map frame ID is included.

    Returns
    -------
    maps : G3Frame or dict of G3Frames
        If ``collate`` is True, returns a dictionary of map frames
        keyed by Id.  Otherwise, returns a single map frame.
    """

    pipe = core.G3Pipeline()
    pipe.Add(core.G3Reader, filename=input_files, track_filename=True)

    # build coadds
    if coadder is None:
        coadder = CoaddMaps(
            map_ids=map_ids,
            output_map_id=output_map_id,
            collate=collate,
            weighted=weighted,
            drop_input_frames=True,
            record_obs_id=record_obs_id,
            keep_outputs=True,
        )
    pipe.Add(coadder)

    # drop metadata frames
    pipe.Add(lambda fr: fr.type == core.G3FrameType.Map)

    if output_file:
        pipe.Add(core.G3Writer, filename=output_file)
    pipe.Run()

    if collate:
        return coadder.coadd_frames
    return coadder.coadd_frame


@core.indexmod
def RebinMaps(frame, scale=1, weighted=True, norm=False):
    """
    Rebin any maps in the input frame into larger pixels by summing scale-x-scale
    blocks of pixels together.  Map dimensions must be a multiple of the rebinning
    scale.

    Arguments
    ---------
    scale : int
        Rebinning scale factor, such that scale-x-scale blocks of pixels are summed
        into larger pixels.  Must divide evenly into all map pixel dimensions.
    weighted : bool
        If True (default), ensure that maps have had weights applied before
        rebinning.  Otherwise, rebin maps without checking the weights.
    norm : bool
        If False (default), sum all of the sub-pixels into each larger output pixel.
        If True, normalize the output pixel by the number of sub-pixels each
        encloses.  Only applies to Stokes parameters T/Q/U, not weights or hits.
    """
    if scale <= 1:
        return

    if weighted:
        ApplyWeights(frame)

    for k in ["T", "Q", "U", "Wunpol", "Wpol", "H"]:
        if k in frame:
            m = frame.pop(k)
            if k in "TQU":
                frame[k] = m.rebin(scale, norm=norm)
            elif k == "H":
                frame[k] = m.rebin(scale, norm=False)
            else:
                if weighted and norm:
                    core.log_warn(
                        "Rebinning weighted frames with norm=True may result "
                        "in inconsistent maps and weights",
                        unit="RebinMaps",
                    )
                frame[k] = m.rebin(scale)


@core.indexmod
class ReprojectMaps(object):
    """
    Reproject a map frame into a different projection.  Original data are
    dropped and replaced by reprojected maps in the input frames.  Maps can be
    changed between flat sky and healpix pixelizations, rotated between
    Equatorial and Galactic coordinates, and/or change polarization convention
    between COSMO and IAU, by setting the appropriate attributes of the input
    and stub maps.  Attributes not defined in the stub map are assumed to be
    that of the input map.  NB: coordinate rotation of polarized maps is not
    currently implemented.

    Arguments
    ---------
    map_stub : G3SkyMap object
        A stub (empty) sky map object to be used to construct the output maps.
        Can be a HealpixSkyMap or FlatSkyMap object.  Setting the ``pol_conv``
        and/or ``coord_ref`` attributes to values that differ from those of the
        input maps will result in output maps whose polarization convention
        and/or reference coordinate system have been changed.
    rebin : int
        If supplied and >1, subdivide the output pixel by n x n with each
        sub-pixel taking on the input map values at pixel center (with interp or
        nearest neighbor). The output pixel takes on the average of the
        sub-pixel values.  In the case that the input map has higher resolution
        than the output map (and that the input map is not low-pass filtered to
        remove information above the Nyquist freq. of the output map pixel),
        this reduces aliasing compared with direct sampling. But there would
        still be aliased power from the input map from freq above the ouput map
        pixel's Nyquist.
    interp : bool
        If True, use bilinear interpolation to extract values from the input
        map.  Otherwise, the nearest-neighbor value is used.
    weighted : bool
        If True (default), ensure that maps have had weights applied before
        reprojection.  Otherwise, reproject maps without checking the weights.
    partial :  bool
        If True, the reproj will be performed on a partial map (of the output map),
        defined by the mask. If the mask is not provided, it will be determined from
        the non-zero pixels of the first reprojected map.
    mask : G3SkyMapMask, G3SkyMap, or np.ndarray, optional
        Mask to be used for partial reproject. This should be of the same size as the
        output map. For numpy array, all zeros/inf/nan/hp.UNSEEN pixels are skipped.
    """

    def __init__(
        self,
        map_stub=None,
        rebin=1,
        interp=False,
        weighted=True,
        partial=False,
        mask=None,
    ):
        assert map_stub is not None, "map_stub argument required"
        self.stub = map_stub.clone(False)
        self.stub.pol_type = None
        self.rebin = rebin
        self.interp = interp
        self.weighted = weighted
        self._mask = None
        self.partial = partial
        self.mask = mask

    def __call__(self, frame):
        if isinstance(frame, core.G3Frame) and frame.type != core.G3FrameType.Map:
            return

        if "Q" in frame and self.stub.coord_ref != frame["Q"].coord_ref:
            raise RuntimeError(
                "Coordinate rotation of polarized maps is not implemented"
            )

        if self.weighted:
            ApplyWeights(frame)

        if "U" in frame and not self.stub.polarized:
            self.stub.pol_conv = frame["U"].pol_conv

        for key in ["T", "Q", "U", "Wpol", "Wunpol", "H"]:

            if key not in frame:
                continue

            m = frame.pop(key)

            if key in "TQUH":
                mnew = self.stub.clone(False)
                maps.reproj_map(
                    m, mnew, rebin=self.rebin, interp=self.interp, mask=self.mask
                )

            elif key in ["Wpol", "Wunpol"]:
                mnew = maps.G3SkyMapWeights(self.stub)
                for wkey in mnew.keys():
                    maps.reproj_map(
                        m[wkey],
                        mnew[wkey],
                        rebin=self.rebin,
                        interp=self.interp,
                        mask=self.mask,
                    )

            frame[key] = mnew
            self.mask = mnew
        return frame

    @property
    def mask(self):
        """
        The mask to be used for partial reprojection, of the same shape as the
        output map.  Masked (1) pixels are handled by the reprojection code, and
        unmasked (0) pixels are skipped, effectively setting their value to 0 in
        the output map.
        """
        return self._mask

    @mask.setter
    def mask(self, mask):
        if mask is None:
            return
        if self._mask is None and self.partial:
            if isinstance(mask, maps.G3SkyMapMask):
                self._mask = mask
            elif isinstance(mask, maps.G3SkyMap):
                self._mask = maps.G3SkyMapMask(
                    mask, use_data=True, zero_nans=True, zero_infs=True
                )
            elif isinstance(mask, np.ndarray):
                import healpy as hp

                tmp = self.stub.clone(False)
                mask_copy = np.ones(mask.shape, dtype=int)
                bad = np.logical_or.reduce(
                    [
                        np.isnan(mask),
                        np.isinf(mask),
                        mask == 0,
                        hp.mask_bad(mask),
                    ]
                )
                mask_copy[bad] = 0
                tmp[:] = mask_copy
                self._mask = maps.G3SkyMapMask(tmp, use_data=True)
            else:
                raise TypeError("Mask must be a G3SkyMapMask, G3SkyMap, or numpy array")

            if not self._mask.compatible(self.stub):
                raise ValueError("Mask is not compatible with output map")



import numpy as np
import re
import datetime as dt
from .. import core
from ..core import G3Units as U
from . import ARCExtractor

def build_field_list(fr):
    """
    All the fields you can upload from gcp to InfluxDB
    """
    r = {
        # tracker status
        'az_actual': ['TrackerStatus', 'az_pos', U.deg],
        'el_actual': ['TrackerStatus', 'el_pos', U.deg],
        'az_rate_actual': ['TrackerStatus', 'az_rate', U.deg/U.sec],
        'el_rate_actual': ['TrackerStatus', 'el_rate', U.deg/U.sec],
        'az_command': ['TrackerStatus', 'az_command', U.deg],
        'el_command': ['TrackerStatus', 'el_command', U.deg],
        'az_rate_command': ['TrackerStatus', 'az_rate_command', U.deg/U.sec],
        'el_rate_command': ['TrackerStatus', 'el_rate_command', U.deg/U.sec],
        'tracker_state': ['TrackerStatus', 'state', None],
        'acu_seq': ['TrackerStatus', 'acu_seq', None],
        'in_control_int': ['TrackerStatus', 'in_control_int', None],
        'scan_flag': ['TrackerStatus', 'scan_flag', None],
        'lst': ['TrackerStatus', 'lst', U.hour],
        'source_acquired': ['TrackerStatus', 'source_acquired', None],
        'source_acquired_thresh': ['TrackerStatus', 'source_acquired_threshold', None],
        'tracker_mode': ['TrackerStatus', 'tracker_mode', None],
        'tracker_lacking': ['TrackerStatus', 'tracker_lacking', None],
        'time_status': ['TrackerStatus', 'time_status', None],
        'schedule': ['TrackerStatus', 'schedule_name', None],
        'raw_encoder_1': ['antenna0', 'tracker', 'raw_encoder', 0, U.deg],
        'raw_encoder_2': ['antenna0', 'tracker', 'raw_encoder', 1, U.deg],
        'drive_currents_el1': ['array', 'dc', 'currents', 0, U.volt],
        'drive_currents_el2': ['array', 'dc', 'currents', 1, U.volt],
        'drive_currents_el3': ['array', 'dc', 'currents', 2, U.volt],
        'drive_currents_el4': ['array', 'dc', 'currents', 3, U.volt],
        'drive_currents_az1': ['array', 'dc', 'currents', 4, U.volt],
        'drive_currents_az2': ['array', 'dc', 'currents', 5, U.volt],
        'drive_currents_az3': ['array', 'dc', 'currents', 6, U.volt],
        'drive_currents_az4': ['array', 'dc', 'currents', 7, U.volt],

        # tracker pointing
        'features': ['TrackerPointing', 'features', 1],
        'encoder_off_x': ['TrackerPointing', 'encoder_off_x', U.deg],
        'encoder_off_y': ['TrackerPointing', 'encoder_off_y', U.deg],
        'low_limit_az': ['TrackerPointing', 'low_limit_az', U.deg],
        'high_limit_az': ['TrackerPointing', 'high_limit_az', U.deg],
        'low_limit_el': ['TrackerPointing', 'low_limit_el', U.deg],
        'high_limit_el': ['TrackerPointing', 'high_limit_el', U.deg],
        'tilts_x': ['TrackerPointing', 'tilts_x', U.deg],
        'tilts_y': ['TrackerPointing', 'tilts_y', U.deg],
        'refraction': ['TrackerPointing', 'refraction', U.deg],
        'horiz_mount_x': ['TrackerPointing', 'horiz_mount_x', U.deg],
        'horiz_mount_y': ['TrackerPointing', 'horiz_mount_y', U.deg],
        'horiz_topo_az': ['TrackerPointing', 'horiz_topo_az', U.deg],
        'horiz_topo_el': ['TrackerPointing', 'horiz_topo_el', U.deg],
        'horiz_off_x': ['TrackerPointing', 'horiz_off_x', U.deg],
        'horiz_off_y': ['TrackerPointing', 'horiz_off_y', U.deg],
        'scan_off_x': ['TrackerPointing', 'scan_off_x', U.deg],
        'scan_off_y': ['TrackerPointing', 'scan_off_y', U.deg],
        'sky_off_x': ['TrackerPointing', 'sky_off_x', U.deg],
        'sky_off_y': ['TrackerPointing', 'sky_off_y', U.deg],
        'equat_off_x': ['TrackerPointing', 'equat_off_x', U.deg],
        'equat_off_y': ['TrackerPointing', 'equat_off_y', U.deg],
        'source_ra': ['TrackerPointing', 'equat_geoc_ra', U.rahr],
        'source_dec': ['TrackerPointing', 'equat_geoc_dec', U.deg],
        'error_az': ['TrackerPointing', 'error_az', U.deg],
        'error_el': ['TrackerPointing', 'error_el', U.deg],
        'linsens_avg_l1': ['TrackerPointing', 'linsens_avg_l1', U.mm],
        'linsens_avg_l2': ['TrackerPointing', 'linsens_avg_l2', U.mm],
        'linsens_avg_r1': ['TrackerPointing', 'linsens_avg_r1', U.mm],
        'linsens_avg_r2': ['TrackerPointing', 'linsens_avg_r2', U.mm],
        'linsens_daz': ['LinearSensorDeltas', 'delta_az', U.deg],
        'linsens_del': ['LinearSensorDeltas', 'delta_el', U.deg],
        'linsens_det': ['LinearSensorDeltas', 'delta_et', U.deg],

        # Weather
        'telescope_temp': ['Weather', 'telescope_temp', 'C'],
        'inside_dsl_temp': ['Weather', 'inside_dsl_temp', 'C'],
        'telescope_pressure': ['Weather', 'telescope_pressure', U.mb],
        'wind_speed': ['Weather', 'wind_speed', U.m / U.s],
        'wind_direction': ['Weather', 'wind_direction', U.deg],
        'battery': ['Weather', 'battery', U.mV],
        'rel_humidity': ['Weather', 'rel_humidity', None],
        'power': ['Weather', 'power', None],
        'tau': ['Weather', 'tau', None],
        'tatm': ['Weather', 'tatm', None],

        # Cryo -- units appear to just be in K. Don't recalibrate.
        # He10
        'uc_head': ['CryoStatus', 'uc_head', 1],
        'ic_head': ['CryoStatus', 'ic_head', 1],
        'he4_head': ['CryoStatus', 'he4_head', 1],
        'he4_fb': ['CryoStatus', 'he4_fb', 1],
        'he4_pump': ['CryoStatus', 'he4_pump', 1],
        'ic_pump': ['CryoStatus', 'ic_pump', 1],
        'uc_pump': ['CryoStatus', 'uc_pump', 1],
        'he4_sw': ['CryoStatus', 'he4_sw', 1],
        'ic_sw': ['CryoStatus', 'ic_sw', 1],
        'uc_sw': ['CryoStatus', 'uc_sw', 1],
        'uc_stage': ['CryoStatus', 'uc_stage', 1],
        'lc_tower': ['CryoStatus', 'lc_tower', 1],
        'ic_stage': ['CryoStatus', 'ic_stage', 1],
        '4k_head': ['CryoStatus', 't4k_head', 1],
        '4k_squid_strap': ['CryoStatus', 't4k_squid_strap', 1],
        '50k_head': ['CryoStatus', 't50k_head', 1],
        # Optics
        'b1_50k_wbp_near': ['CryoStatus', 'b1_50k_wbp_near', 1],
        'b2_50k_wbp_far': ['CryoStatus', 'b2_50k_wbp_far', 1],
        'b3_50k_diving_board': ['CryoStatus', 'b3_50k_diving_board', 1],
        'b4_50k_top_bot_ptc': ['CryoStatus', 'b4_50k_top_bot_ptc', 1],
        'y1_50k_head': ['CryoStatus', 'y1_50k_head', 1],
        'y2_50k_window_strap_near': ['CryoStatus', 'y2_50k_window_strap_near', 1],
        'y3_50k_tube_strap_near': ['CryoStatus', 'y3_50k_tube_strap_near', 1],
        'y4_50k_tube': ['CryoStatus', 'y4_50k_tube', 1],
        'g1_4k_head': ['CryoStatus', 'g1_4k_head', 1],
        'g2_4k_strap': ['CryoStatus', 'g2_4k_strap', 1],
        'g3_4k_lens_tab': ['CryoStatus', 'g3_4k_lens_tab', 1],
        'g4_4k_lens_tab_far': ['CryoStatus', 'g4_4k_lens_tab_far', 1],
        'r1_4k_top_top_ptc': ['CryoStatus', 'r1_4k_top_top_ptc', 1],
        'r2_50k_midop_bot_ptc': ['CryoStatus', 'r2_50k_midop_bot_ptc', 1],
        'r3_4k_lyot_flange': ['CryoStatus', 'r3_4k_lyot_flange', 1],
        'r4_4k_lyot': ['CryoStatus', 'r4_4k_lyot', 1],
        # Receiver
        '4k_plate_far': ['CryoStatus', 't4k_plate_far', 1],
        '4k_strap_optics': ['CryoStatus', 't4k_strap_optics', 1],
        '4k_plate_mid': ['CryoStatus', 't4k_plate_mid', 1],
        '4k_plate_top': ['CryoStatus', 't4k_plate_top', 1],
        '4k_plate_ptc': ['CryoStatus', 't4k_plate_ptc', 1],
        '50k_harness_middle': ['CryoStatus', 't50k_harness_middle', 1],
        '50k_strap': ['CryoStatus', 't50k_strap', 1],
        'squid_wh1_sl1': ['CryoStatus', 'squid_wh1_sl1', 1],
        'squid_wh5_sl1': ['CryoStatus', 'squid_wh5_sl1', 1],
        'squid_wh3_sl7': ['CryoStatus', 'squid_wh3_sl7', 1],
        'cal_filament': ['CryoStatus', 'cal_filament', 1],
        'cal_ambient1': ['CryoStatus', 'cal_ambient1', 1],
        'cal_ambient2': ['CryoStatus', 'cal_ambient2', 1],
        'cal_ambient3': ['CryoStatus', 'cal_ambient3', 1],
        # heaters
        'heat_he4_pump': ['CryoStatus', 'heat_he4_pump', 1],
        'heat_ic_pump': ['CryoStatus', 'heat_ic_pump', 1],
        'heat_uc_pump': ['CryoStatus', 'heat_uc_pump', 1],
        'heat_he4_sw': ['CryoStatus', 'heat_he4_sw', 1],
        'heat_ic_sw': ['CryoStatus', 'heat_ic_sw', 1],
        'heat_uc_sw': ['CryoStatus', 'heat_uc_sw', 1],
        # status bit
        'cryo_is_valid': ['CryoStatus', 'cryo_is_valid', None],

        # PT status
        'optics_low_p_now': ['PTStatus', 'optics_lowp', None],
        'optics_low_p_min': ['PTStatus', 'min_optics_lowp', None],
        'optics_low_p_max': ['PTStatus', 'max_optics_lowp', None],
        'optics_high_p_now': ['PTStatus', 'optics_highp', None],
        'optics_high_p_min': ['PTStatus', 'min_optics_highp', None],
        'optics_high_p_max': ['PTStatus', 'max_optics_highp', None],
        'optics_tempoil_now': ['PTStatus', 'optics_tempoil', None],
        'optics_tempoil_min': ['PTStatus', 'min_optics_tempoil', None],
        'optics_tempoil_max': ['PTStatus', 'max_optics_tempoil', None],

        'receiver_low_p_now': ['PTStatus', 'receiver_lowp', None],
        'receiver_low_p_min': ['PTStatus', 'min_receiver_lowp', None],
        'receiver_low_p_max': ['PTStatus', 'max_receiver_lowp', None],
        'receiver_high_p_now': ['PTStatus', 'receiver_highp', None],
        'receiver_high_p_min': ['PTStatus', 'min_receiver_highp', None],
        'receiver_high_p_max': ['PTStatus', 'max_receiver_highp', None],
        'receiver_tempoil_now': ['PTStatus', 'receiver_tempoil', None],
        'receiver_tempoil_min': ['PTStatus', 'min_receiver_tempoil', None],
        'receiver_tempoil_max': ['PTStatus', 'max_receiver_tempoil', None],

        'optics_is_valid': ['PTStatus', 'optics_is_valid', None],
        'receiver_is_valid': ['PTStatus', 'receiver_is_valid', None],

        # Online Pointing Model
        'tilts_hr_angle': ['OnlinePointingModel', 'tilts', 0, U.deg],
        'tilts_lat': ['OnlinePointingModel', 'tilts', 1, U.deg],
        'tilts_el': ['OnlinePointingModel', 'tilts', 2, U.deg],
        'flexure_sin': ['OnlinePointingModel', 'flexure', 0, U.deg],
        'flexure_cos': ['OnlinePointingModel', 'flexure', 1, U.deg],
        'fixed_collimation_x': ['OnlinePointingModel', 'fixedCollimation', 0, U.deg],
        'fixed_collimation_y': ['OnlinePointingModel', 'fixedCollimation', 1, U.deg],
        'tilts_hr_angle_adjust': ['OnlinePointingModelCorrection', 'tilts', 0, U.deg],
        'tilts_lat_adjust': ['OnlinePointingModelCorrection', 'tilts', 1, U.deg],
        'tilts_el_adjust': ['OnlinePointingModelCorrection', 'tilts', 2, U.deg],
        'flexure_sin_adjust': ['OnlinePointingModelCorrection', 'flexure', 0, U.deg],
        'flexure_cos_adjust': ['OnlinePointingModelCorrection', 'flexure', 1, U.deg],
        'fixed_collimation_x_adjust': ['OnlinePointingModelCorrection', 'fixedCollimation', 0, U.deg],
        'fixed_collimation_y_adjust': ['OnlinePointingModelCorrection', 'fixedCollimation', 1, U.deg],
        'linsens_coeff_az': ['OnlinePointingModel', 'linsensCoeffs', 0, None],
        'linsens_coeff_el': ['OnlinePointingModel', 'linsensCoeffs', 1, None],
        'linsens_coeff_et': ['OnlinePointingModel', 'linsensCoeffs', 2, None],
        'linsens_enabled': ['OnlinePointingModel', 'linsensEnabled', 0, None],

        # Other
        'obs_id': ['ObservationID', None],
        'source_name': ['SourceName', None],

        # ACUStatus
        'acu_state': ['ACUStatus', 'state', None],
        'acu_status': ['ACUStatus', 'status', None],
        'acu_error': ['ACUStatus', 'error', None],

        # Bench
        'bench_command_y1': ['BenchCommandedPosition', 'y1', U.mm],
        'bench_command_y2': ['BenchCommandedPosition', 'y2', U.mm],
        'bench_command_y3': ['BenchCommandedPosition', 'y3', U.mm],
        'bench_command_x4': ['BenchCommandedPosition', 'x4', U.mm],
        'bench_command_x5': ['BenchCommandedPosition', 'x5', U.mm],
        'bench_command_z6': ['BenchCommandedPosition', 'z6', U.mm],

        'bench_actual_y1': ['BenchPosition', 'y1', U.mm],
        'bench_actual_y2': ['BenchPosition', 'y2', U.mm],
        'bench_actual_y3': ['BenchPosition', 'y3', U.mm],
        'bench_actual_x4': ['BenchPosition', 'x4', U.mm],
        'bench_actual_x5': ['BenchPosition', 'x5', U.mm],
        'bench_actual_z6': ['BenchPosition', 'z6', U.mm],

        'bench_zero_y1': ['BenchZeros', 'y1', U.mm],
        'bench_zero_y2': ['BenchZeros', 'y2', U.mm],
        'bench_zero_y3': ['BenchZeros', 'y3', U.mm],
        'bench_zero_x4': ['BenchZeros', 'x4', U.mm],
        'bench_zero_x5': ['BenchZeros', 'x5', U.mm],
        'bench_zero_z6': ['BenchZeros', 'z6', U.mm],

        'bench_offset_y1': ['BenchOffsets', 'y1', U.mm],
        'bench_offset_y2': ['BenchOffsets', 'y2', U.mm],
        'bench_offset_y3': ['BenchOffsets', 'y3', U.mm],
        'bench_offset_x4': ['BenchOffsets', 'x4', U.mm],
        'bench_offset_x5': ['BenchOffsets', 'x5', U.mm],
        'bench_offset_z6': ['BenchOffsets', 'z6', U.mm],

        'bench_error_y1': ['BenchErrors', 'y1', U.mm],
        'bench_error_y2': ['BenchErrors', 'y2', U.mm],
        'bench_error_y3': ['BenchErrors', 'y3', U.mm],
        'bench_error_x4': ['BenchErrors', 'x4', U.mm],
        'bench_error_x5': ['BenchErrors', 'x5', U.mm],
        'bench_error_z6': ['BenchErrors', 'z6', U.mm],

        'bench_focus': ['BenchInfo', 'benchFocus', U.mm],
        'bench_dead_band': ['BenchInfo', 'benchDeadBand', U.mm],
        'bench_acquired_thresh': ['BenchInfo', 'benchAcquiredThreshold', U.mm],
        'bench_primary_state': ['BenchInfo', 'benchPrimaryState', None],
        'bench_secondary_state': ['BenchInfo', 'benchSecondaryState', None],
        'bench_fault': ['BenchInfo', 'benchFault', None],
        'bench_time_locked': ['BenchInfo', 'timeLocked', None],
    }

    # mux housekeeping
    for i in range(32):
        i = str(i)
        r['fpga_temp_ib{}'.format(i)] = ['MuxFPGATemp', i, None]
        r['name_ib{}'.format(i)] = ['MuxBoardName', i, None]

    # scu.temp - all temps documented given a name, others just a number
    scu_temps = {
        0: 'yoke_air',
        1: 'ctrl_room_air',
        2: 'glycol_supply',
        3: 'glycol_return',
        4: 'ctrl_room',
        20: 'secondary',
        21: 'icecrate',
        22: 'bench',
        23: 'attic',
        24: 'cabin',
        25: 'cryoboard',
    }
    for i in range(60):
        key = 't_scu_{}'.format(scu_temps.get(i, i))
        r[key] = ['TrackerPointing', 'scu_temp', i, 'C']

    return r


def make_lines(measurement, field, time, dat, tags=None):
    '''
    Return list of string lines to add to database
    '''
    if isinstance(dat[0], str):
        dat = ['"' + x + '"' for x in dat]
    if tags is None:
        fmt_str = measuremnt +' '+ field + '={val} {time}'
    else:
        fmt_str = measurement
        for tag, tag_val in tags.items():
            fmt_str += ',{}={}'.format(tag, tag_val)
        fmt_str += ' ' + field + '={val} {time}'
    # time in int because it's in nanoseconds for InfluxDB
    lines = [fmt_str.format(val=x, time=int(t0)) for x, t0 in zip(dat, time)]
    return lines


@core.indexmod
def WriteDB(fr, client, fields=None):
    '''
    Write points to the database for each field

    Arguments
    ---------
    client :
        InfluxDB client
    fields :
        Which gcp fields to add to database. See parse_field for options. If
        None, add all.
    '''
    assert client is not None, "InfluxDB client required"
    from influxdb.exceptions import InfluxDBClientError
    from influxdb.exceptions import InfluxDBServerError

    if fr.type != core.G3FrameType.GcpSlow:
        return
    all_fields = build_field_list(fr)
    if fields is None:
        fields = all_fields.keys()
    dict_list = []
    for f in fields:
        field_dat = all_fields[f]
        if len(field_dat) == 5:
            # raw register
            stat = 'TrackerStatus'
            tmp, brd, attr, ind, unit = field_dat
            try:
                dat = fr[tmp][brd][attr][ind]
            except:
                continue
            try:
                if 'utc' in fr[tmp][brd].keys():
                    time = fr[tmp][brd]['utc'][0]
                else:
                    time = fr['antenna0']['tracker']['utc'][0][:len(dat)]
            except:
                continue
        elif len(field_dat) == 4:
            stat, attr, ind, unit = field_dat
            if stat not in fr:
                # OnlinePointingModelCorrection
                continue
            try:
                dat = getattr(fr[stat], attr)[ind]
                time = getattr(fr[stat], 'time')
            except AttributeError:
                # OnlinePointingModel
                try:
                    dat = fr[stat][attr][ind]
                except KeyError:
                    # OnlinePointingModelCorrection
                    continue
                time = fr[stat]['time']
        elif len(field_dat) == 3:
            stat, attr, unit = field_dat
            if stat not in fr:
                # Field only exists in live data stream
                continue
            try:
                dat = getattr(fr[stat], attr)
            except AttributeError:
                try:
                    dat = fr[stat][attr]
                except KeyError: # Field only exists in live data stream
                    continue
            if 'Bench' in stat: # funny time field for bench positions
                time = fr['BenchSampleTime']
            elif 'Mux' in stat:
                time = fr['MuxTime']
            elif stat in ['CryoStatus', 'Weather', 'PTStatus']:
                time = fr['{}Time'.format(stat)]
            else:
                try:
                    time = getattr(fr[stat], 'time')
                except AttributeError:
                    time = fr[stat]['time']
        elif len(field_dat) == 2:
            stat, unit = field_dat
            try:
                dat = fr[stat]
            except KeyError: #eg, no obsid
                core.log_warn('No key {}'.format(stat), unit='InfluxDB')
                continue
            try:
                time = getattr(fr[stat], 'time')
            except AttributeError as err:
                time = fr['antenna0']['tracker']['utc'][0]

        # InfluxDB wants time in nanoseconds since the UNIX epoch in UTC
        if isinstance(time, core.G3Time):
            time = np.asarray([time.time / U.nanosecond])
        elif isinstance(time, core.G3VectorTime):
            time = np.asarray(time) / U.nanosecond
        else:
            try:
                time = np.asarray(core.G3VectorTime(np.atleast_1d(time))) / U.nanosecond
            except Exception as e:
                core.log_error("Error converting time: {}".format(str(e)), unit="InfluxDB")
                continue
        # Avoid out-of-range errors in parsing time
        # time < 0 occurs when, e.g., the SCU loses the clock (e.g. during a brownout)
        time[time < 0] = 0
        if dat is None:
            core.log_warn('{} dat is None'.format(f), unit='InfluxDB')
            continue
        dat = np.atleast_1d(dat)
        try:
            dlen = len(dat)
        except TypeError:
            # sometimes source_name is a weird non-none value
            continue
        if unit is not None:
            if unit == 'C':
                zeropt_K = 273.15
                cal_dat = dat / U.K - zeropt_K
            else:
                cal_dat = dat / unit
        else:
            cal_dat = dat
        try:
            if np.any(np.isnan(cal_dat)):
                continue
        except TypeError:
            pass
        if 'heat' not in f:
            tag = f
        else:
            tag = f.replace('heat_', '')

        # for fields that have az/el components
        az_el_names = ['az', 'el', 'az', 'el', 'ra', 'dec', 'x', 'y',
                       'hr_angle', 'sin', 'cos', 'lat']
        tag2 = f
        for name in az_el_names:
            # require name_ at beginning or _name at end
            match1 = re.findall('^{}_'.format(name), f)
            match2 = re.findall('_{}$'.format(name), f)
            if len(match1):
                tag2 = f.replace(match1[0], '')
            if len(match2):
                tag2 = f.replace(match2[0], '')
        # also group source names
        if 'source' in f:
            tag2 = 'source'
            stat = 'TrackerPointing'
        if stat == 'PTStatus':
            groups = ['now', 'min', 'max']
            for g in groups:
                match = re.findall('_{}$'.format(g), f)
                if len(match):
                    tag2 = f.replace(match[0], '')
        # group bench positions
        # require bench_ at beginning
        match = re.findall('^bench', f)
        if len(match):
            tag2 = attr # y1, y2, etc
            stat = 'Bench'

        # group Mux properties
        if 'Mux' in stat:
            stat = 'muxHousekeeping'
            tag2 = 'ib'+f.split('ib')[-1]

        if 'linsens_coeff' in f:
            tag2 = 'linsens_coeff'

        if 'drive_currents_az' in f:
            tag2 = 'drive_currents_az'
        if 'drive_currents_el' in f:
            tag2 = 'drive_currents_el'

        dict_list += make_lines(
            measurement=stat,
            field=f,
            time=time,
            dat=cal_dat,
            tags={'label': tag, 'label2': tag2},
        )

    try:
        now = core.G3Time.Now()
        delay = float(now.time/U.nanosecond - time[-1])/1e9
        if delay > 5:
            core.log_info(
                '{} Delay: {} s'.format(now.isoformat(), delay), unit='InfluxDB'
            )
    except RuntimeError: # sometimes timestamp gets screwed up
        pass

    try:
        client.write_points(
            dict_list, batch_size=len(dict_list), protocol='line'
        )
    except (InfluxDBClientError, InfluxDBServerError) as v:
        core.log_error(
            'Error writing to database. {}'.format(v), unit='InfluxDB'
        )


@core.pipesegment
def UpdateDB(pipe, client=None, fields=None):
    '''
    Update InfluxDB with data in frame

    Arguments
    ---------
    client :
        InfluxDB client
    fields :
        Which gcp fields to add to database. See parse_field for options. If
        None, add all.
    '''
    pipe.Add(ARCExtractor.ARCExtract)
    pipe.Add(WriteDB, client=client, fields=fields)



from spt3g.core import G3MapInt
import unittest

class G3MapTestSuite(unittest.TestCase):
    def setUp(self):
        self.d = {"a": 1, "b": 2, "c": 3}
        self.m = G3MapInt(self.d)

    def test_getitem(self):
        self.assertEqual(self.m["a"], 1)
        self.assertEqual(self.m["b"], 2)
        self.assertEqual(self.m["c"], 3)
        with self.assertRaises(KeyError):
            _ = self.m["d"]

    def test_setitem(self):
        self.m["a"] = 10
        self.assertEqual(self.m["a"], 10)
        self.m["d"] = 4
        self.assertEqual(self.m["d"], 4)

    def test_delitem(self):
        del self.m["b"]
        with self.assertRaises(KeyError):
            _ = self.m["b"]
        self.assertEqual(len(self.m), 2)

        del self.m["a"]
        del self.m["c"]
        self.assertTrue(not self.m)

        with self.assertRaises(KeyError):
            del self.m["a"]

    def test_contains(self):
        self.assertTrue("a" in self.m)
        self.assertTrue("b" in self.m)
        self.assertTrue("c" in self.m)
        self.assertFalse("d" in self.m)

    def test_len(self):
        self.assertEqual(len(self.m), 3)
        self.m["d"] = 4
        self.assertEqual(len(self.m), 4)
        del self.m["a"]
        self.assertEqual(len(self.m), 3)

    def test_iter(self):
        keys = sorted(list(self.m.keys()))
        expected_keys = sorted(list(self.d.keys()))
        self.assertEqual(keys, expected_keys)
        
        values = []
        for key in self.m:
            values.append(self.m[key])
        self.assertEqual(sorted(values), sorted(self.d.values()))

    def test_keys(self):
        keys_view = self.m.keys()
        self.assertEqual(sorted(list(keys_view)), sorted(list(self.d.keys())))
        self.assertEqual(len(keys_view), len(self.d))
        self.m["d"] = 4
        self.assertEqual(sorted(list(keys_view)), sorted(list(self.d.keys()) + ["d"]))

    def test_values(self):
        values_view = self.m.values()
        self.assertEqual(sorted(list(values_view)), sorted(list(self.d.values())))
        self.assertEqual(len(values_view), len(self.d))
        self.m["a"] = 10
        values = list(values_view)
        values.sort()
        expected_values = list(self.d.values())
        expected_values[0] = 10
        expected_values.sort()
        self.assertEqual(sorted(values), sorted(expected_values))

    def test_items(self):
        items_view = self.m.items()
        self.assertEqual(sorted(list(items_view)), sorted(list(self.d.items())))
        self.assertEqual(len(items_view), len(self.d))
        self.m["d"] = 4
        self.assertEqual(sorted(list(items_view)), sorted(list(self.d.items()) + [("d", 4)]))

    def test_clear(self):
        self.m.clear()
        self.assertEqual(len(self.m), 0)
        self.assertFalse(self.m)
        self.assertEqual(list(self.m.keys()), [])
        self.assertEqual(list(self.m.values()), [])
        self.assertEqual(list(self.m.items()), [])

    def test_copy(self):
        copied_dict = self.m.copy()
        self.assertEqual(len(copied_dict), len(self.m))
        self.assertEqual(sorted(list(copied_dict.items())), sorted(list(self.m.items())))
        copied_dict["a"] = 100
        self.assertNotEqual(self.m["a"], 100)
        self.assertEqual(copied_dict["a"], 100)

    def test_get(self):
        self.assertEqual(self.m.get("a"), 1)
        self.assertEqual(self.m.get("d"), None)
        self.assertEqual(self.m.get("d", 5), 5)
        self.assertEqual(self.m.get("b", 10), 2)

    def test_pop(self):
        self.assertEqual(self.m.pop("b"), 2)
        self.assertEqual(len(self.m), 2)
        with self.assertRaises(KeyError):
            _ = self.m["b"]
        self.assertEqual(self.m.pop("d", 5), 5)
        with self.assertRaises(KeyError):
            self.m.pop("d")

    def test_update(self):
        self.m.update({"b": 20, "d": 4})
        self.assertEqual(self.m["b"], 20)
        self.assertEqual(self.m["d"], 4)
        self.assertEqual(len(self.m), 4)

        self.m.update([("c", 30), ("e", 5)])
        self.assertEqual(self.m["c"], 30)
        self.assertEqual(self.m["e"], 5)
        self.assertEqual(len(self.m), 5)

        self.m.update(f=6)
        self.assertEqual(self.m["f"], 6)
        self.assertEqual(len(self.m), 6)
        
        self.m.update({})
        self.assertEqual(len(self.m), 6)

        self.m.update([])
        self.assertEqual(len(self.m), 6)
        
        self.m.update()
        self.assertEqual(len(self.m), 6)
    
    def test_bool(self):
        self.assertTrue(self.m)
        self.m.clear()
        self.assertFalse(self.m)

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

# Test that various styles of return values do the right thing in C++.
# Empty frames can confuse Boost, which will trigger exceptions on the C++
# side (hence the lack of assert(), etc. here)

from spt3g import core

class Source(object):
    def __init__(self):
        self.trig = True
    def __call__(self, frame):
        if self.trig:
            self.trig = False
            f = core.G3Frame(core.G3FrameType.Timepoint)
            f['Foo'] = 5
            return f
        return []

class Returner(object):
    def __init__(self):
        return
    def __call__(self, frame):
        print(1, frame)
        return 


class Returner2(object):
    def __init__(self):
        return
    def __call__(self, frame):
        print(2, frame.__class__, frame)
        return frame

p = core.G3Pipeline()
p.Add(Source)
p.Add(Returner)
p.Add(Returner2)
p.Run()



from spt3g import core
from spt3g.calibration import BolometerProperties, set_band_format, band_to_value, get_band_units

bp = BolometerProperties()
bp.band = 94.67 * core.G3Units.GHz

assert bp.band_string == "95GHz"
assert bp.band_vstring == "95"
assert band_to_value(bp.band_string) == 95 * core.G3Units.GHz
assert bp.band_vstring + get_band_units() == bp.band_string

set_band_format(0, "MHz")

assert bp.band_string == "94670MHz"
assert bp.band_vstring == "94670"
assert band_to_value(bp.band_string) == bp.band
assert bp.band_vstring + get_band_units() == bp.band_string

set_band_format(2, "GHz")

assert bp.band_string == "94.67GHz"
assert bp.band_vstring == "94.67"
assert band_to_value(bp.band_string) == bp.band
assert bp.band_vstring + get_band_units() == bp.band_string

set_band_format(-1, "GHz")
assert bp.band_string == "90GHz"
assert bp.band_vstring == "90"
assert band_to_value(bp.band_string) == 90 * core.G3Units.GHz
assert bp.band_vstring + get_band_units() == bp.band_string



from .. import core

__all__ = ["G3File"]

@core.usefulfunc
class G3File(object):
    """Iterable class for G3 files, as created by G3Writer. Loop through frames by doing something like::

        with core.G3File("/path/to/file.g3") as f:
            for frame in f:
                print(frame)

    An entire file can also be read into an indexable list by doing::

        f = list(core.G3File("/path/to/file.g3"))
    """

    def __init__(self, path):
        self.reader = core.G3Reader(path)

    def __iter__(self):
        return self

    def next(self):
        """
        Return the next frame from the file.
        """
        if self.reader is None:
            raise StopIteration("Reader closed")
        frames = self.reader.Process(None)
        if len(frames) == 0:
            raise StopIteration("No more frames in file")
        if len(frames) > 1:
            raise ValueError("Too many frames returned by reader")
        return frames[0]

    __next__ = next

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        del self.reader
        self.reader = None


# writer context methods


def writer_enter(self):
    return self


def writer_exit(self, *args, **kwargs):
    fr = core.G3Frame(core.G3FrameType.EndProcessing)
    self(fr)


core.G3Writer.__enter__ = writer_enter
core.G3Writer.__exit__ = writer_exit
core.G3MultiFileWriter.__enter__ = writer_enter
core.G3MultiFileWriter.__exit__ = writer_exit



import logging
import os
import re
import subprocess
import sys
import sysconfig
from pathlib import Path

from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext
from setuptools.command.install import install
from setuptools.command.install_scripts import install_scripts
from setuptools.command.sdist import sdist


# A CMakeExtension does not need a source list
class CMakeExtension(Extension):
    def __init__(self, name):
        super().__init__(name, sources=[])


class CMakeBuildExt(build_ext):
    def initialize_options(self):
        super().initialize_options()
        self.source_dir = Path(".").resolve()
        self.build_dir = Path(os.getenv("BUILD_DIR", "./build")).resolve()

        cmake_args = []

        # Adding CMake arguments set as environment variable
        if "CMAKE_ARGS" in os.environ:
            cmake_args += [item for item in os.environ["CMAKE_ARGS"].split(" ") if item]

        # sensible defaults
        if not any(["Python_ROOT_DIR" in a for a in cmake_args]):
            pyroot = sysconfig.get_config_var("prefix")
            cmake_args += [f"-DPython_ROOT_DIR={pyroot}"]
        if not any(["Python_EXECUTABLE" in a for a in cmake_args]):
            cmake_args += [f"-DPython_EXECUTABLE={sys.executable}"]
        if not any(["pybind11_DIR" in a for a in cmake_args]):
            import pybind11
            pbroot = pybind11.get_cmake_dir()
            cmake_args += [f"-Dpybind11_DIR={pbroot}"]

        # install libraries and headers into virtual environment too
        if "VIRTUAL_ENV" in os.environ:
            if not any(["CMAKE_INSTALL_PREFIX" in a for a in cmake_args]):
                envroot = os.getenv("VIRTUAL_ENV")
                cmake_args += [f"-DCMAKE_INSTALL_PREFIX={envroot}"]

        # pass version to C++ code
        if not any(["SPT3G_VERSION" in a for a in cmake_args]):
            cmake_args += [f"-DPIP_SPT3G_VERSION_FILE=ON"]
            version = self.distribution.metadata.version
            if re.match(r"^[0-9]+\.[0-9]+$", version):
                cmake_args += [f"-DSPT3G_VERSION={repr(version)}"]

        self.cmake_args = cmake_args
        self.cmake_bin = os.getenv("CMAKE_BIN", "cmake")
        self.cmake_done = False

    def build_extension(self, ext):
        if not self.cmake_done:
            # build once
            self.announce(f"Building library in {self.build_dir}", logging.INFO)
            if not self.build_dir.exists():
                self.build_dir.mkdir(parents=True, exist_ok=True)
            subprocess.run(
                [self.cmake_bin, self.source_dir, *self.cmake_args], cwd=self.build_dir, check=True
            )
            subprocess.run([self.cmake_bin, "--build", "."], cwd=self.build_dir, check=True)

            # update package directory
            if self.editable_mode:
                pkgdir = self.build_dir.relative_to(self.source_dir) / "spt3g"
                self.distribution.package_dir["spt3g"] = pkgdir
                build_py = self.get_finalized_command("build_py")
                build_py.package_dir["spt3g"] = pkgdir

            # update version file
            self.copy_file(self.source_dir / "cmake/package/version.py", self.build_dir / "spt3g")

            # trigger script installer for all python scripts
            self.distribution.scripts = [
                str(f) for f in self.build_dir.glob("bin/*") if f.is_symlink()
            ]
            self.cmake_done = True

        # add modules
        spt3g_lib = Path(self.build_lib) / "spt3g"
        if not spt3g_lib.exists():
            spt3g_lib.mkdir(parents=True)

        libname = ext.name.split(".")[-1]
        libfiles = [
            self.build_dir / "spt3g" / self.get_ext_filename(libname),
            self.build_dir / "spt3g" / f"{libname}.so",
        ]
        for f in libfiles:
            if f.exists():
                self.copy_file(f, spt3g_lib)
                break
        else:
            raise RuntimeError(f"Missing extension module {ext.name}")


class CMakeInstallScripts(install_scripts):
    def run(self):
        super().run()

        self.announce("Installing cmake programs", logging.INFO)
        install_dir = Path(self.install_dir)
        if not install_dir.exists():
            install_dir.mkdir(parents=True)

        for src in self.get_finalized_command("build_ext").build_dir.glob("bin/*"):
            self.copy_file(src, install_dir)


class CMakeInstall(install):
    def run(self):
        super().run()

        if "CIBUILDWHEEL" in os.environ:
            return

        self.announce("Installing cmake libraries and headers", logging.INFO)
        build_dir = self.get_finalized_command("build_ext").build_dir
        subprocess.run(["make", "install"], cwd=build_dir, check=True)


class ArchiveDist(sdist):
    def run(self):

        if Path(".git").exists():
            ar = subprocess.run(
                ["git", "archive", "HEAD", ".git_archival.txt"],
                check=True,
                capture_output=True,
            )
            subprocess.run(["tar", "-x"], input=ar.stdout, check=True)

        super().run()

        if Path(".git").exists():
            subprocess.run(["git", "checkout", ".git_archival.txt"], check=True)


# gather libraries
clibs = []
pdirs = {"spt3g": "cmake/package"}

for d in sorted(Path("./").glob("*/CMakeLists.txt")):
    lib = d.parent.name
    if (d.parent / "src").exists():
        clibs.append(f"spt3g._lib{lib}")
    if (d.parent / "python").exists():
        pdirs[f"spt3g.{lib}"] = d.parent / "python"
    elif (d.parent / "__init__.py").exists():
        pdirs[f"spt3g.{lib}"] = d.parent

setup(
    ext_modules=[CMakeExtension(lib) for lib in clibs],
    cmdclass={
        "build_ext": CMakeBuildExt,
        "install_scripts": CMakeInstallScripts,
        "install": CMakeInstall,
        "sdist": ArchiveDist,
    },
    packages=list(pdirs),
    package_dir=pdirs,
    exclude_package_data={k: ["CMakeLists.txt"] for k in pdirs},
)



from .. import core

@core.indexmod
class MPIAccumulator(object):
    '''
    Accumulate data from many frames in memory, sharing metadata for all the
    frames between nodes in an MPI communicator, potentially organized into
    groups (for example by observation ID).

    The result will be stored in the member variable 'fullobs' after
    pipeline termination. 'fullobs' is a dictionary containing entries
    for each data group (see the extractfunc parameter), which in turn
    are lists of 3-element tuples with the following content:

    - First element is the metadata for the frame (see extractfunc)
    - Second element is the node on which the full data exist
    - Third element is the data from the frame (see extractfunc)
      or None if the data are on a remote process.
    '''
    def __init__(self, mpicomm, extractfunc=None, sorter=None, dataframes=[core.G3FrameType.Scan, core.G3FrameType.Timepoint]):
        '''
        Data will be shared between nodes participating in the communicator
        mpicomm. The data stored are based on frames of the types in the
        dataframes argument and are organized based on the results of
        ``extractfunc`` and ``sorter``.

        ``extractfunc`` is expected to return a three-element tuple:

        - First element is a group ID for the data (for example, an observation
          ID). This is used only for organization and can be anything that
          Python can compare. By default, this is SourceName-ObservationID.
        - Second element is metadata. This information is shared between
          all processes participating in the communicator at completion and
          should include any information you need to assess whether the data
          are interesting for future calculations (for example, the start time
          of the frame). By default, frames are sorted by this value. None
          by default.
        - Third element is the data. This stays node-local. It could be just
          a copy of the frame (the default) or just the information (in any
          format understandable to Python) from the frame you expect to use
          later.

        Note that ``extractfunc`` is called on all frame types (including
        calibration data) but *only the results if it returns non-None values*
        are stored. Thus, it is the responsibility of ``extractfunc`` to either
        cache any applicable calibration, etc. information and store it in the
        return value for data frames if needed or explicitly return a value
        for calibration data, if needed. The default ``extractfunc`` saves only
        the frames in the "dataframes" argument.

        ``sorter`` is used, if defined, to sort the list of frame data in each
        group. This is passed to the Python sorted() function as a 'key'
        argument and receives the three-element tuple stored for each set of
        frame data (metadata, node, data). By default, does no sorting
        (frames will still appear in a common order across nodes related to
        the data distribution if unsorted).
        '''
        self.mpicomm = mpicomm
        self.extractfunc = extractfunc
        def defaultextract(f):
            if f.type not in self.dataframes:
                return
            return ('%s-%d' % (f['SourceName'], f['ObservationID']), None, core.G3Frame(f))
        if self.extractfunc is None:
            self.extractfunc = defaultextract
        self.sorter = sorter
        self.localdata = {}
    def __call__(self, frame):
        if frame.type == core.G3FrameType.EndProcessing:
            self.finalize()
            return

        extracted = self.extractfunc(frame)
        if extracted is None:
            return

        obskey, metadata, data = extracted
        if obskey not in self.localdata:
            self.localdata[obskey] = []
        self.localdata[obskey].append((metadata, data))

    def finalize(self):
        # Put all observation metadata everywhere
        fullobs = {}
        localchunk = {}
        for k,v in self.localdata.items():
            localchunk[k] = [(m, self.mpicomm.rank, None) for m,d in v]
        chunks = self.mpicomm.allgather(localchunk)
        del localchunk
        for k,v in self.localdata.items():
            chunks[self.mpicomm.rank][k] = [(m, self.mpicomm.rank, d) for m,d in v]
        del self.localdata
        for chunk in chunks:
            for obs, data in chunk.items():
                if obs not in fullobs:
                    fullobs[obs] = []
                fullobs[obs] += data
        del chunks
            
        # Now let's organize it
        if self.sorter is not None:
            for obs, data in fullobs.items():
                data.sort(key=self.sorter)

        self.fullobs = fullobs




from . import ARCFileReader, ARCExtract
from spt3g import core

@core.usefulfunc
class ARCFile(object):
    '''Iterable class for ARC files, as created by GCP. Loop through frames by doing something like::

        f = gcp.ARCFile('/path/to/arc.dat')
        for frame in f:
            print( frame )

    An entire file can also be read into an indexable list by doing::

        f = list(gcp.ARCFile('/path/to/arc.dat'))
    '''
    def __init__(self, path, extract=False):
        self.reader = ARCFileReader(path)
        self.extract = extract

    def __iter__(self):
        return self

    def next(self):
        frames = self.reader.Process(None)
        if len(frames) == 0:
            raise StopIteration('No more frames in file')
        if len(frames) > 1:
            raise ValueError('Too many frames returned by reader')
        frame = frames[0]

        if self.extract:
            # calibrate and parse arc frames
            from ..core import G3Pipeline, G3InfiniteSource, InjectFrame

            pipe = G3Pipeline()
            pipe.Add(G3InfiniteSource, n=1)
            pipe.Add(InjectFrame, frame=frame)
            pipe.Add(ARCExtract)
            pipe.Run()

        return frame

    __next__ = next



#!/usr/bin/env python

from spt3g import core
import time

nframes = 1000
verbose = False

# File to disk
core.log_notice("Writing %d frames to disk" % nframes)
pipe = core.G3Pipeline()
pipe.Add(core.G3InfiniteSource, type=core.G3FrameType.Timepoint, n=nframes)
# Drop a wiring frame in the middle
count = 0
def addwiring(fr):
    global count
    if fr.type == core.G3FrameType.Timepoint:
        count += 1
        if count == 6:
            fr2 = core.G3Frame(core.G3FrameType.Wiring)
            return [fr2, fr]
pipe.Add(addwiring)
n = 0
def addinfo(fr):
    global n
    if fr.type != core.G3FrameType.Timepoint:
        return
    fr['time'] = core.G3Time(int(time.time()*core.G3Units.s))
    fr['count'] = n
    n += 1
pipe.Add(addinfo)
if verbose:
    pipe.Add(core.Dump)
pipe.Add(core.G3Writer, filename='test.g3')
pipe.Run(profile=True)
assert n == nframes, 'Wrong number of frames written (%d should be %d)' % (n, nframes)

# And back from disk
core.log_notice("Reading %d frames from disk" % nframes)
pipe = core.G3Pipeline()
pipe.Add(core.G3Reader, filename='test.g3', track_filename=True)
if verbose:
    pipe.Add(core.Dump)
n = 0
def checkinfo(fr):
    global n
    if fr.type != core.G3FrameType.Timepoint:
        return
    assert 'time' in fr, 'No time key in frame'
    assert fr['count'] == n, 'Out of order frame'
    assert fr._filename == 'test.g3', 'Wrong filename'
    n += 1
pipe.Add(checkinfo)
pipe.Run(profile=True)

assert n == nframes, 'Wrong number of frames read (%d should be %d)' % (n, nframes)

# Skip empty files
core.log_notice("Reading %d frames from disk with empty files" % nframes)
with core.G3Writer("empty.g3") as wr:
    pass
n = 0
pipe = core.G3Pipeline()
pipe.Add(core.G3Reader, filename=["empty.g3", "test.g3", "empty.g3"], track_filename=True)
pipe.Add(checkinfo)
pipe.Run(profile=True)

assert n == nframes, 'Wrong number of frames read (%d should be %d)' % (n, nframes)

# Indexing
class CachingReader:
    def __init__(self, filename='test.g3'):
        self.reader = core.G3Reader(filename='test.g3')
        self.w_pos = None

    def __call__(self, frame):
        assert frame is None
        pos = self.reader.tell()
        fr = self.reader(frame)
        if not len(fr):
            return fr
        if fr[0].type == core.G3FrameType.Wiring:
            self.w_pos = pos
        return fr

cacher = CachingReader()

core.log_notice("Reading %d frames from disk with frame indexing" % nframes)
pipe = core.G3Pipeline()
pipe.Add(cacher)
if verbose:
    pipe.Add(core.Dump)
pipe.Run(profile=True)

assert cacher.w_pos is not None, 'Missing wiring frame'

# Using cached index
class CachedReader:
    def __init__(self, filename='test.g3', start=None):
        self.reader = core.G3Reader(filename=filename)
        self.pos = start

    def __call__(self, frame):
        assert frame is None
        if self.pos is not None:
            self.reader.seek(self.pos)
        if self.pos is not None:
            assert self.reader.tell() == self.pos
        fr = self.reader(frame)
        if not len(fr):
            return fr
        if self.pos is not None:
            assert fr[0].type == core.G3FrameType.Wiring
            self.pos = None
        return fr

cached = CachedReader(start=cacher.w_pos)

core.log_notice("Reading %d frames from disk with seek" % nframes)
pipe = core.G3Pipeline()
pipe.Add(cached)
if verbose:
    pipe.Add(core.Dump)
pipe.Run(profile=True)

assert cached.pos is None, "Missing wiring frame"

# check EOF handling
while cached.reader(None):
    pass

# can seek to EOF if at EOF
pos = cached.reader.tell()
cached.reader.seek(pos)

# can't seek backward once at EOF
try:
    cached.reader.seek(0)
except:
    pass
else:
    assert False, "Expected RuntimeError seeking back from EOF"

# check append mode
core.log_notice("Testing writer in append mode")
with core.G3Writer("appended.g3") as wr:
    fr = core.G3Frame(core.G3FrameType.Wiring)
    wr(fr)

with core.G3Writer("appended.g3", append=True) as wr:
    assert wr.tell() > 0, "File pointer not moved to the end"
    fr = core.G3Frame(core.G3FrameType.Timepoint)
    wr(fr)

frs = list(core.G3File("appended.g3"))
assert len(frs) == 2, "Missing frames in append mode"



#!/usr/bin/env python

import numpy
from spt3g import core, maps
from spt3g.maps import FlatSkyMap, MapCoordReference, MapProjection

# 1D arrays
m = FlatSkyMap(1, 500, core.G3Units.arcmin)
assert(m.shape[0] == 500)
m[15] = 65.4
assert(numpy.asarray(m)[15] == 65.4)
assert(numpy.asarray(m).shape == m.shape)

# 2D arrays
m = FlatSkyMap(500, 20, core.G3Units.arcmin)
assert(len(m.shape) == 2)
assert(m.shape[0] == 20)
assert(m.shape[1] == 500)
assert(m.shape[0] == numpy.asarray(m).shape[0])
assert(m.shape[1] == numpy.asarray(m).shape[1])
m[15] = 65.4
assert(numpy.asarray(m)[0,15] == 65.4)
m[3,15] = 65.4
assert(numpy.asarray(m)[3,15] == 65.4)
assert(numpy.asarray(m).shape == m.shape)

# Reverse direction 2D arrays
w = numpy.random.uniform(size=(300,50))
m = FlatSkyMap(w, core.G3Units.arcmin, coord_ref=MapCoordReference.Equatorial)
assert(len(m.shape) == 2)
assert(m.shape[0] == w.shape[0])
assert(m.shape[1] == w.shape[1])
assert(w[17,32] == m[17,32])
assert((w == m).all())

# Test conversion between representations is lossless
# Along the way, test nonzero_pixels() and rebin() once for each
# representation

a = numpy.arange(1500, dtype=float)
a[0] = -1
v = numpy.zeros((50, 50), dtype=float)
v.ravel()[:1500] = a
x = FlatSkyMap(v, core.G3Units.arcmin, proj=MapProjection.ProjZEA)

try:
    v2 = numpy.asarray(x)[10:20, 10:20] # non-contiguous array
    x2 = FlatSkyMap(v2, core.G3Units.arcmin)
    numpy.testing.assert_array_equal(x2, v2)
except Exception as e:
    print(str(e))
    pass
else:
    raise RuntimeError("Expected error!")

assert(not x.sparse)
assert(x.npix_allocated == x.size)
assert(x[1499] == 1499)
assert(x[1501] == 0)

k, v = x.nonzero_pixels()
assert(len(k) == len(v) == 1500)
assert(set(v) == set(a))

v0 = x[0, 0] + x[0, 1] + x[1, 0] + x[1, 1]

x2 = x.rebin(2, norm=False)
assert(x2[0] == v0)
assert(numpy.sum(x2) == numpy.sum(v))
assert(numpy.allclose(x.xy_to_angle(-0.5, -0.5), x2.xy_to_angle(-0.5, -0.5)))

x.sparse = True
assert(x.npix_allocated == 1500)
for i in range(1, 1500):
    assert(x[i] == i)
assert(x[1501] == 0)

k, v = x.nonzero_pixels()
assert(len(k) == len(v) == 1500)
assert(set(v) == set(a))

x2 = x.rebin(2, norm=False)
assert(x2[0] == v0)
assert(numpy.sum(x2) == numpy.sum(v))

x.sparse = False
assert(x.npix_allocated == x.size)
for i in range(1, 1500):
    assert(x[i] == i)
assert(x[1501] == 0)

k, v = x.nonzero_pixels()
assert(len(k) == len(v) == 1500)
assert(set(v) == set(a))

x2 = x.rebin(2, norm=False)
assert(x2[0] == v0)
assert(numpy.sum(x2) == numpy.sum(v))

# query disc
alpha, delta = maps.get_ra_dec_map(x)

from astropy.coordinates import SkyCoord
from astropy import units as u

angles = SkyCoord(
    numpy.asarray(alpha) / core.G3Units.deg * u.degree,
    numpy.asarray(delta) / core.G3Units.deg * u.degree,
).ravel()
radius = 5 * x.res

avec = []
dvec = []
rvec = []
masked = set()

for a in numpy.linspace(numpy.min(alpha), numpy.max(alpha), 20):
    for d in numpy.linspace(numpy.min(delta), numpy.max(delta), 20):
        avec.append(a)
        dvec.append(d)
        rvec.append(radius)
        pix1 = x.query_disc(a, d, radius)
        pix2 = numpy.where(
            angles.separation(
                SkyCoord(
                    a / core.G3Units.deg * u.degree,
                    d / core.G3Units.deg * u.degree,
                )
            ).deg * core.G3Units.deg < radius
        )[0]
        masked |= set(pix2)
        assert not (set(pix1) ^ set(pix2))

mask = maps.make_point_source_mask(x, numpy.asarray(avec), numpy.asarray(dvec), numpy.asarray(rvec))
assert not (set(mask.nonzero()) ^ masked)



import numpy
from .. import core
from .Housekeeping import HousekeepingForBolo

# Transfer functions for 3G and SPTpol boards
from .IceboardConversions import convert_TF, convert_adc_samples
from .wtl_ConvertUtils import convert_squid, convert_demod, convert_mb

def counts_to_rms_amps(wiringmap, hkmap, bolo, system, tf=None):
    boardhk, mezzhk, modhk, chanhk = HousekeepingForBolo(hkmap, wiringmap, bolo, True)

    if system == 'ICE':
        if chanhk.dan_streaming_enable:
            tf_I = convert_TF(modhk.nuller_gain, target='nuller', custom_TF=tf, unit='RAW', frequency=chanhk.carrier_frequency/core.G3Units.Hz)
        else:
            tf_V = convert_adc_samples('Streamer')
            tf_I = tf_V/modhk.squid_transimpedance
        if chanhk.carrier_frequency != 0:
            tf_I /= numpy.sqrt(2) # Use RMS amps
    elif system == 'DfMux':
        if chanhk.dan_streaming_enable:
            squid_converter = convert_squid(units='Normalized')
            tf_I = 1. / (2.**24) * squid_converter.theo_AnToIbolo(1., Gnuller=modhk.nuller_gain, eff=1.)
        else:
            dfmux_converter = convert_demod('gcp')
            resistor='10K' # Assume this for DAN Off
            # Handle the three exceptions for SPTpol
            if bolo in ['Sq3SBpol23Ch7', 'Sq3SBpol23Ch3', 'Sq3SBpol23Ch5']:
                resistor='5K'
            # XXX above list is wrong for 100d data
            if boardhk.timestamp.mjd < 56293:
                core.log_fatal('Set correct resistor values for 2012 SPTpol data', unit='dfmux.unittransforms')
            tf_I = dfmux_converter.theo_AdcToIbolo(1., modhk.demod_gain, resistor, f_carrier_equals_f_demod=True)
            tf_I *= 256 # 3G software shifts 8 bits right compared to SPTpol, so account for difference in meaning of "counts"
    else:
        core.log_fatal('Unknown transfer function profile %s' % system, unit='dfmux.unittransforms')

    return tf_I * core.G3Units.amp

def bolo_bias_voltage_rms(wiringmap, hkmap, bolo, system, tf=None):
    boardhk, mezzhk, modhk, chanhk = HousekeepingForBolo(hkmap, wiringmap, bolo, True)

    if system == 'ICE':
        tf_V = convert_TF(modhk.carrier_gain, target='carrier', custom_TF=tf, unit='NORMALIZED', frequency=chanhk.carrier_frequency/core.G3Units.Hz)
        volts = tf_V * chanhk.carrier_amplitude * core.G3Units.V
        if chanhk.carrier_frequency != 0:
            volts /= numpy.sqrt(2) # Use RMS amps
    elif system == 'DfMux':
        squid_converter = convert_squid(units='Normalized')
        volts = squid_converter.theo_AcToVbias(chanhk.carrier_amplitude, modhk.carrier_gain) * core.G3Units.V
    else:
        core.log_fatal('Unknown transfer function profile %s' % system, unit='dfmux.unittransforms')

    return volts

@core.usefulfunc
def get_timestream_unit_conversion(from_units, to_units, bolo, wiringmap=None, hkmap=None, system=None, tf=None):
    '''
    Return the scalar conversion factor to move timestream data from a
    given system of units (Power, Resistance, Current, Counts) to another one.
    Requires a wiring map and recent housekeeping data.
    Returned quantities are RMS for currents and time-averaged for power.

    Note that this does not handle conversions to on-sky quantities (e.g. K_cmb)
    '''

    # Rather than special-casing the full mesh of conversions,
    # first calculate a conversion to watts from whatever units
    # we have, then a conversion from watts

    if wiringmap is None:
        core.log_fatal('Try to convert data with no wiring map', unit='dfmux.unittransforms')
    if hkmap is None:
        core.log_fatal('Try to convert data with no housekeeping', unit='dfmux.unittransforms')

    if from_units == to_units:
        return 1.

    if system not in ["ICE", "DfMux"]:
        core.log_fatal(f"Cannot convert data for {system} readout system", unit="dfmux.unittransforms")

    I = counts_to_rms_amps(wiringmap, hkmap, bolo, system, tf)
    V = bolo_bias_voltage_rms(wiringmap, hkmap, bolo, system, tf)

    if from_units == core.G3TimestreamUnits.Resistance:
        if to_units == core.G3TimestreamUnits.Counts:
            return V / I
        if to_units == core.G3TimestreamUnits.Current:
            return V
        if to_units == core.G3TimestreamUnits.Power:
            return V * V
    if to_units == core.G3TimestreamUnits.Resistance:
        if from_units == core.G3TimestreamUnits.Counts:
            return V / I
        if from_units == core.G3TimestreamUnits.Current:
            return V
        if from_units == core.G3TimestreamUnits.Power:
            return V * V

    # First, convert to watts
    if from_units == core.G3TimestreamUnits.Counts:
        to_watts = I * V
    elif from_units == core.G3TimestreamUnits.Current:
        to_watts = V
    elif from_units == core.G3TimestreamUnits.Power:
        to_watts = 1.
    else:
        core.log_fatal('Unknown units scheme %s in timestream for bolo %s' % (repr(from_units), bolo), unit='dfmux.unittransforms')

    # Now the conversion from watts
    if to_units == core.G3TimestreamUnits.Counts:
        from_watts = I * V
        if from_watts != 0:
            from_watts = 1./from_watts
    elif to_units == core.G3TimestreamUnits.Current:
        from_watts = V
        if from_watts != 0:
            from_watts = 1./from_watts
    elif to_units == core.G3TimestreamUnits.Power:
        from_watts = 1.
    else:
        core.log_fatal('Trying to convert timestreams to unknown units %s' % repr(to_units), unit='dfmux.unittransforms')

    # Multiply them together
    return to_watts * from_watts

@core.indexmod
class ConvertTimestreamUnits(object):
    '''
    Changes timestream units from one set of units (e.g. ADC counts) to
    another (e.g. Power). Currents and power are time averaged quantities
    (i.e. currents give RMS values).

    Note that this does not handle conversions to on-sky quantities (e.g. K_cmb)
    '''

    def __init__(self, Input='RawTimestreams', Output='CalTimestreams', Units=core.G3TimestreamUnits.Power, SkipUncalibratable=False, KeepConversionsForObservation=True):
        '''
        Copy data in the timestream map Input to the timestream map Output,
        converting the units from whatever they were to those specified by
        Units.

        If SkipUncalibratable is true, copy timestreams for which the
        unit conversions could not be evaluated into the output timestream
        in their original units. If false, throws an exception if this occurs.

        If KeepConversionsForObservation is True (default), conversion factors
        will only be evaluated once per observation. Note that this will cause
        the wrong conversions to be applied if any parameters of the mux system
        are modified during the observation but will substantially increase
        performance.
        '''
        self.units = Units
        self.input = Input
        self.output = Output

        self.wiringmap = None
        self.system = None
        self.default_tf = None
        self.hkmap = None
        self.convfactors = {}
        self.skiperrors = SkipUncalibratable
        self.keepconversions = KeepConversionsForObservation

    def __call__(self, frame):
        if 'DfMuxTransferFunction' in frame:
            self.default_tf = frame['DfMuxTransferFunction']
        if frame.type == core.G3FrameType.Wiring:
            self.wiringmap = frame['WiringMap']
            self.system = frame['ReadoutSystem']
            self.convfactors = {}
            return
        if frame.type == core.G3FrameType.Housekeeping:
            self.hkmap = frame['DfMuxHousekeeping']
            self.convfactors = {}
            return
        if self.keepconversions and frame.type == core.G3FrameType.Observation:
            self.convfactors = {}
            self.hkmap = None
            return

        if frame.type != core.G3FrameType.Scan:
            return

        # Housekeeping data can also be in Scan frames
        if 'DfMuxHousekeeping' in frame:
            if self.hkmap is None or (list(frame['DfMuxHousekeeping'].values())[0].timestamp != list(self.hkmap.values())[0].timestamp and not self.keepconversions):
                # XXX: finer-grained check for same values?
                self.hkmap = frame['DfMuxHousekeeping']
                self.convfactors = {}

        # Now the meat
        newts = core.G3TimestreamMap()
        oldts = frame[self.input]
        for bolo,ts in oldts.items():
            if ts.units not in self.convfactors:
                self.convfactors[ts.units] = {}
            if ts.units == self.units:
                newts[bolo] = ts
                continue
            elif bolo in self.convfactors[ts.units]:
                convfactor = self.convfactors[ts.units][bolo]
            else:
                tf = self.default_tf # XXX: might get from HK data
                try:
                    convfactor = get_timestream_unit_conversion(ts.units, self.units, bolo, wiringmap=self.wiringmap, hkmap=self.hkmap, system=self.system, tf=tf)
                except KeyError:
                    if not self.skiperrors:
                        raise
                    newts[bolo] = ts
                    continue
                self.convfactors[ts.units][bolo] = convfactor # And cache it

            # Convert timestream and store results
            convts = core.G3Timestream(ts)
            if core.G3TimestreamUnits.Resistance in [self.units, ts.units] and self.units != ts.units:
                convts = 1. / convts
            convts.units = self.units
            convts *= convfactor
            if convts.units != core.G3TimestreamUnits.Counts:
                convts.SetFLACCompression(False)
            newts[bolo] = convts
        frame[self.output] = newts




#!/usr/bin/env python

from __future__ import print_function

import numpy, sys
from spt3g import core

then = core.G3Time.Now()
now = core.G3Time(then.time + 3*core.G3Units.second)

data = numpy.zeros(200)

ts = core.G3Timestream(data)
ts.start = then
ts.stop = now

assert(numpy.abs(ts.sample_rate/core.G3Units.Hz - 66.33333) < 1e-5)

times = ts.times

assert(ts.times[0] == ts.start)
print(ts.times[-1].time, ts.stop.time)
assert(numpy.abs(ts.times[-1].time - ts.stop.time) < 2) # Max 1 tick roundoff error

tsm = core.G3TimestreamMap()
tsm['Test'] = ts

assert(numpy.abs(tsm.sample_rate/core.G3Units.Hz - 66.33333) < 1e-5)

times = tsm.times

assert(tsm.times[0] == tsm.start)
assert(numpy.abs(tsm.times[-1].time - tsm.stop.time) < 2) # Max 1 tick roundoff error




#!/usr/bin/env python

from spt3g import core
import sys

def AddInfo(frame):
    frame['Info'] = 12

# First, make sure the check doesn't trip when it isn't supposed to. 
pipe = core.G3Pipeline()
pipe.Add(core.G3InfiniteSource, type=core.G3FrameType.Timepoint, n=10)
pipe.Add(AddInfo)
pipe.Add(lambda f: 1/0)
try:
    pipe.Run() # Will throw an exception if test fails
except:
    pass

if (pipe.last_frame['Info'] != 12):
    sys.exit(1)

pipe = core.G3Pipeline()

if not pipe.last_frame is None:
    sys.exit(1)



import numpy as np
from spt3g.core import (
    G3Timestream,
    G3TimestreamMap,
    G3VectorDouble,
    DoubleVector,
    FloatVector,
    G3VectorInt,
    Int64Vector,
    IntVector,
    G3VectorComplexDouble,
    ComplexDoubleVector,
    ComplexFloatVector,
    G3VectorBool,
    BoolVector,
)

s = 5.0  # python float
si = 5  # python integer
sc = s + 1j * s
v = np.abs(np.random.randn(10)) + 1.0  # numpy array
vi = np.random.randint(1, 10, 10)
vc = v + 1j * v  # numpy complex array
f = v[0]  # numpy float
fi = vi[0]  # numpy integer
fc = vc[0]  # numpy complex float
sb = True
vb = v.astype(bool)
fb = vb[0]

all_arr = [
    s,
    f,
    v,
    si,
    fi,
    vi,
    sc,
    fc,
    vc,
    sb,
    fb,
    vb,
    G3Timestream(v),
    G3Timestream(vi),
    G3VectorDouble(v),
    DoubleVector(v),
    FloatVector(v),
    G3VectorInt(vi),
    Int64Vector(vi),
    IntVector(vi),
    G3VectorComplexDouble(vc),
    ComplexDoubleVector(vc),
    ComplexFloatVector(vc),
    G3VectorBool(vb),
    BoolVector(G3VectorBool(vb)),
]

def check(a1, a2, cls=None, flt=False, short=False):
    if flt:
        np.testing.assert_allclose(
            np.asarray(a1),
            np.asarray(a2),
            atol=1e-4 if short else 1e-14,
            rtol=1e-4 if short else 1e-14,
            verbose=True,
        )
    else:
        np.testing.assert_array_equal(np.asarray(a1), np.asarray(a2), verbose=True)
    assert np.asarray(a1).dtype.kind == np.asarray(a2).dtype.kind
    if cls is not None:
        assert isinstance(a1, cls)

def copy(v):
    if isinstance(v, np.ndarray):
        return v.copy()
    return v.__class__(v)

for a in all_arr:
    for b in all_arr:
        n = "/".join([a.__class__.__name__, b.__class__.__name__])
        if not ("Vector" in n or "Timestream" in n):
            continue

        print(type(a), type(b), type(a + b))
        v1 = a if np.isscalar(a) else np.asarray(a)
        v2 = b if np.isscalar(b) else np.asarray(b)
        k1 = np.asarray(a).dtype.kind
        k2 = np.asarray(b).dtype.kind
        flt = k1 in 'fc' or k2 in 'fc'
        c1 = np.asarray(a).dtype.char
        c2 = np.asarray(b).dtype.char
        short = c1 in 'fF' or c2 in 'fF'

        # binary operators
        if not (k1 == 'b' and k2 == 'b'):
            check(a + b, v1 + v2, flt=flt, short=short)
            check(a - b, v1 - v2, flt=flt, short=short)
            check(a * b, v1 * v2, flt=flt, short=short)
            check(a / b, v1 / v2, flt=flt, short=short)
            if k1 != 'c' and k2 != 'c':
                check(a // b, v1 // v2)
                check(a % b, v1 % v2, flt=flt, short=short)
            check(a ** b, v1 ** v2, flt=flt, short=short)

        if k1 in 'iub' and k2 in 'iub':
            check(a | b, v1 | v2)
            check(a & b, v1 & v2)
            check(a ^ b, v1 ^ v2)
            check(a << 1, v1 << 1)
            check(a >> 1, v1 >> 1)

        cls = None if np.isscalar(a) or isinstance(a, np.ndarray) else G3VectorBool if 'G3' in n else BoolVector
        check(a < b, v1 < v2, cls)
        check(a <= b, v1 <= v2, cls)
        check(a == b, v1 == v2, cls)
        check(a != b, v1 != v2, cls)
        check(a >= b, v1 >= v2, cls)
        check(a > b, v1 > v2, cls)

        if np.isscalar(a) or k1 == 'b':
            continue

        korder = list('buifc')
        if korder.index(np.asarray(a).dtype.kind) < korder.index(np.asarray(b).dtype.kind):
            continue

        # in-place binary operators
        c = copy(a)
        c += b
        check(c, a + b, a.__class__, flt=flt, short=short)

        c = copy(a)
        c -= b
        check(c, a - b, a.__class__, flt=flt, short=short)

        c = copy(a)
        c *= b
        check(c, a * b, a.__class__, flt=flt, short=short)

        if k1 in 'iub':
            c = copy(a)
            c |= b
            check(c, a | b, a.__class__)

            c = copy(a)
            c &= b
            check(c, a & b, a.__class__)

            c = copy(a)
            c ^= b
            check(c, a ^ b, a.__class__)

            continue

        c = copy(a)
        c /= b
        check(c, a / b, a.__class__, flt=flt, short=short)

    if np.isscalar(a) or isinstance(a, np.ndarray):
        continue

    for attr in [
        "sum",
        "mean",
        "any",
        "all",
        "min",
        "max",
        "argmin",
        "argmax",
        "var",
        "std",
    ]:
        assert getattr(a, attr)() == getattr(np, attr)(v1)

    if isinstance(a, G3Timestream):
        tsm = G3TimestreamMap(["a", "b", "c"], np.tile(v1, (3, 1)))

        for attr in ["std", "var"]:
            np.testing.assert_allclose(
                getattr(tsm, attr)(axis=-1),
                getattr(np, attr)(np.asarray(tsm), axis=-1),
                atol=1e-14,
                rtol=1e-14,
                verbose=True,
            )



#!/usr/bin/env python

# Test that the checks for EndProcessing frames in G3Pipeline work.

from spt3g import core
import sys

# First, make sure the check doesn't trip when it isn't supposed to.
pipe = core.G3Pipeline()
pipe.Add(core.G3InfiniteSource, type=core.G3FrameType.Timepoint, n=10)
pipe.Add(lambda f: True)
pipe.Run() # Will throw an exception if test fails

# First, make sure the check doesn't trip when it isn't supposed to.
pipe = core.G3Pipeline()
pipe.Add(core.G3InfiniteSource, type=core.G3FrameType.Timepoint, n=10)
pipe.Add(lambda f: False)
pipe.Run() # Will throw an exception if test fails

# Next, make sure it fails if EndProcessing frames are dropped
pipe = core.G3Pipeline()
pipe.Add(core.G3InfiniteSource, type=core.G3FrameType.Timepoint, n=10)
pipe.Add(lambda f: [])
try:
    pipe.Run()
except RuntimeError:
    print('Exception thrown on all frames dropped - good!')
    pass
else:
    print('No exception when EndProcessing frame dropped')
    sys.exit(1)
	
# A subtler test: make sure EndProcessing is always the *last* frame
pipe = core.G3Pipeline()
pipe.Add(core.G3InfiniteSource, type=core.G3FrameType.Timepoint, n=10)
pipe.Add(lambda f: [f, core.G3Frame(core.G3FrameType.Scan)])
try:
    pipe.Run()
except RuntimeError:
    print('Exception thrown when EndProcessing followed by scan - good!')
    pass
else:
    print('No exception when EndProcessing followed by scan')
    sys.exit(1)
	
print('EndProcessing tests pass!')



import spt3g
import spt3g._libcore
for k in dir(spt3g._libcore):
    print(k, getattr(spt3g._libcore, k))
from spt3g._libcore import BoolVector
import spt3g.core
from spt3g import core
from spt3g.core import BoolVector



#!/usr/bin/env python

from spt3g import core, dfmux
import socket, argparse, fnmatch

parser = argparse.ArgumentParser(description='Record dfmux data to a NetCDF file based on connection to a G3NetworkSender tee in a running DAQ script', prog='ledgerman')
parser.add_argument('output', metavar='output.nc', help='Path to output NetCDF file')
parser.add_argument('server', metavar='host.name', help='Server running data acquisition')
parser.add_argument('port', metavar='port', default=5353, help='Port of DAQ server')

parser.add_argument('-m', dest='hardware_map', default=None,
                    help='Path to hardware map YAML file. Uses the network version if unspecified.')

parser.add_argument('-v', dest='verbose', action='store_true',
                    help='Verbose mode (print all frames)')
parser.add_argument('-P', dest='physnames', action='store_true',
                    help='Use physical bolometer names rather than channel names')
parser.add_argument('-p', dest='pathstring', action='store', default=[], nargs='+',
                    help='Only record channels that match channel path string,'
                    ' e.g. with a local pydfmux hardware map, 005/5/2/3/*'
                    ' saves data for all bolometers on crate 005, slot 5, '
                    'mezzanine 2, module 3. When used with a remote hardware '
                    'map, applied as a glob to the stored bolometer names.')
parser.add_argument('-b', dest='state', action='store', default=[], nargs='+',
                    help='Only record bolometers that have a state that'
                    ' matches the supplied string(s), e.g. overbiased tuned')
args = parser.parse_args()


if args.hardware_map is not None:
    # Import pydfmux later since it can take a while
    import pydfmux

    core.log_notice('Initializing hardware map and boards', unit='TeeLedgerman')
    hwm = pydfmux.load_session(open(args.hardware_map, 'r'))['hardware_map']
    if hwm.query(pydfmux.IceCrate).count() > 0:
        hwm.query(pydfmux.IceCrate).resolve()

    # Retrieve pydfmux state
    # Make sure the hardware map is consistent with what's on the IceBoards
    if args.state:
        hwm.query(pydfmux.Bolometer).load_bolo_states()

core.log_notice('Beginning data acquisition', unit='TeeLedgerman')

# Set up DfMux consumer
pipe = core.G3Pipeline()
pipe.Add(core.G3Reader, filename='tcp://' + args.server + ':' + str(args.port))

if args.hardware_map is not None:
    # Insert current hardware map, replacing whatever we got from the server,
    # if user requests it
    pipe.Add(lambda fr: fr.type != core.G3FrameType.Wiring)
    pipe.Add(dfmux.PyDfMuxHardwareMapInjector, pydfmux_hwm=hwm)

    if args.physnames:
        # Need bolo properties below
        pipe.Add(dfmux.PyDfMuxBolometerPropertiesInjector, pydfmux_hwm=hwm)

if args.physnames:
    from spt3g import calibration
    # Swap the normal bolometer names for their physical_name equivalents,
    # which can make laboratory tasks simpler -- we don't care about
    # long-term data archiving here.
    class SwapPhysLogicalNames(object):
        def __init__(self):
            self.calframe = None
            self.wiringframe = None
            self.sent = False

        def __call__(self, frame):
            if self.sent:
                return
            if frame.type == core.G3FrameType.Wiring:
                self.wiringframe = frame
                return []
            elif frame.type == core.G3FrameType.Calibration:
                self.calframe = frame
                return []
            elif frame.type != core.G3FrameType.Timepoint:
                # Just pass this along...
                return

            if self.calframe is None or self.wiringframe is None:
                raise Exception('Data before wiring and cal frames!')
            w = self.wiringframe['WiringMap']
            del self.wiringframe['WiringMap']
            c = self.calframe['NominalBolometerProperties']

            new_wiring = dfmux.DfMuxWiringMap()
            for k, d in w.iteritems():
                new_wiring[c[k].physical_name] = d
            self.wiringframe['WiringMap'] = new_wiring

            self.sent = True
            return [self.wiringframe, frame]
    pipe.Add(SwapPhysLogicalNames)

if len(args.pathstring) > 0:
    def trim_wiring_map(fr):
        if fr.type != core.G3FrameType.Wiring:
            return

        fr['OriginalWiringMap'] = fr['WiringMap']
        del fr['WiringMap']
        keys_to_keep = set()
        for item in args.pathstring:
            keys_to_keep |= set(fnmatch.filter(fr['OriginalWiringMap'].keys(), item))
        new_wiring = dfmux.DfMuxWiringMap()
        for k in keys_to_keep:
            new_wiring[k] = fr['OriginalWiringMap'][k]
        fr['WiringMap'] = new_wiring
        core.log_notice('Taking data from %d/%d detectors' % (len(keys_to_keep), len(fr['OriginalWiringMap'])), unit='TeeLedgerman')
    pipe.Add(trim_wiring_map)

if args.verbose:
    pipe.Add(core.Dump)

pipe.Add(dfmux.NetCDFDump, filename=args.output)
pipe.Run()




from spt3g import core, dfmux, networkstreamer, hk

pipe = core.G3Pipeline()
pipe.Add(networkstreamer.G3NetworkReceiver, 
         hostname = 'laphroaig.berkeley.edu', 
         port = 8675)



pipe.Add(core.Dump)

pipe.Run()



#!/usr/bin/env python

import numpy, sys
from spt3g import core

ts = core.G3Timestream(numpy.ones(1200))

tsarr = numpy.asarray(ts)

if tsarr[5] != ts[5]:
	print('Values corrupted')
	sys.exit(1)

if len(tsarr) != len(ts):
	print('Lengths do not match')
	sys.exit(1)

tsarr[6] = 15
if ts[6] != 15:
	print('Buffer not shared')
	sys.exit(1)

ts[7] = 45
if tsarr[7] != 45:
	print('Buffer not shared')
	sys.exit(1)

print(tsarr[5:9])

sys.exit(0)




from .. import core, maps

import numpy as np

__all__ = ["flatsky_to_healpix", "healpix_to_flatsky"]


@core.usefulfunc
def healpix_to_flatsky(
    map_in, nest=False, map_stub=None, rebin=1, interp=False, **kwargs
):
    """
    Re-pixelize a map from Healpix to one of the flat sky projections.

    All additional keyword arguments are passed to FlatSkyMap to construct the
    output map object.  Required if `map_stub` is not supplied, otherwise
    ignored.

    Maps can be rotated between Equatorial and Galactic coordinates, and/or
    change polarization convention between COSMO and IAU, by setting the
    appropriate attributes of the input and output maps.  Note that if the input
    map is a numpy array representation of a healpix map, the coordinate system
    and polarization convention are assumed to be that of the output map.
    Conversely, attributes not defined in the output map are assumed to be
    that of the input map.

    Parameters
    ----------
    map_in: numpy array or HealpixSkyMap
        The array containing the input healpix map to reproject.

    nest: bool
        Ordering of the healpix map, if the input is a numpy array.  Ring
        ordering is assumed by default.

    map_stub: FlatSkyMap
        Stub output map object to be used to construct the output map.  If not
        supplied, one will be constructed using the remaining keyword arguments.

    rebin: int
        If supplied and >1, subdivide the output pixel by n x n with each 
        sub-pixel taking on the input map values at pixel center (with interp
        or nearest neighbor). The output pixel takes on the average of the sub-pixel
        values.
        In the case that the input map has higher resolution than the output
        map (and that the input map is not low-pass filtered to remove
        information above the Nyquist freq. of the output map pixel), this
        reduces aliasing compared with direct sampling. But there would still be
        aliased power from the input map from freq above the ouput map pixel's
        Nyquist. 

    interp: bool
        If True, use bilinear interpolation to extract values from the 4 closest
        pixel centers of the healpix map. Otherwise, the nearest-neighbor value 
        is used.

    Returns
    -------
    output_map: FlatSkyMap
        The reprojected map
    """

    # Construct the output map
    if map_stub is None:
        if len(kwargs) == 0:
            raise ValueError("Need to provide either a stub map or map parameters (projection, center, etc.)")
        if isinstance(map_in, maps.HealpixSkyMap):
            kwargs.setdefault("coord_ref", map_in.coord_ref)
            kwargs.setdefault("pol_type", map_in.pol_type)
            kwargs.setdefault("pol_conv", map_in.pol_conv)
        map_out = maps.FlatSkyMap(**kwargs)
    else:
        if not isinstance(map_stub, maps.FlatSkyMap):
            raise TypeError("Output stub must be a FlatSkyMap")
        map_out = map_stub.clone(False)

    # Populate output map pixels with interpolation and rebinning
    if not isinstance(map_in, maps.HealpixSkyMap):
        map_in = maps.HealpixSkyMap(
            map_in,
            nested=nest,
            coord_ref=map_out.coord_ref,
            weighted=map_out.weighted,
            units=map_out.units,
            pol_type=map_out.pol_type,
            pol_conv=map_out.pol_conv,
        )
    maps.reproj_map(map_in, map_out, rebin=rebin, interp=interp)

    return map_out


@core.usefulfunc
def flatsky_to_healpix(
    map_in, map_stub=None, rebin=1, interp=False, fullsky=False, **kwargs
):
    """
    Re-pixelize a map to Healpix from one of the flat projections.

    All additional keyword arguments are passed to HealpixSkyMap to construct
    the output map object.  Required if `map_stub` is not supplied,
    otherwise ignored.

    Maps can be rotated between Equatorial and Galactic coordinates, and/or
    change polarization convention between COSMO and IAU, by setting the
    appropriate attributes of the input and output maps.  Attributes not defined
    in the output map are assumed to be that of the input map.

    Parameters
    ----------
    map_in: FlatSkyMap
        The input map you want to reproject

    map_stub: HealpixSkyMap
        Stub output map object to be used to construct the output map.  If not
        supplied, one will be constructed using the remaining keyword arguments.

    rebin: int
        If supplied and >1, subdivide the output pixel by n x n with each 
        sub-pixel taking on the input map values at pixel center (with interp
        or nearest neighbor). The output pixel takes on the average of the sub-pixel
        values.
        In the case that the input map has higher resolution than the output
        map (and that the input map is not low-pass filtered to remove
        information above the Nyquist freq. of the output map pixel), this
        reduces aliasing compared with direct sampling. But there would still be
        aliased power from the input map from freq above the ouput map pixel's
        Nyquist. 

    interp: bool
        If True, use bilinear interpolation to extract values from the input
        map.  Otherwise, the nearest-neighbor value is used.

    fullsky: bool
        If True a full-sky numpy array representation of the map is returned.
        Otherwise, a HealpixSkyMap instance is returned, containing only the
        pixels that overlap with the input map.

    Returns
    -------
    output_map: numpy array or HealpixSkyMap
        The array containing the healpix map.  If `fullsky` is True, this is a
        numpy array, otherwise a HealpixSkyMap instance.
    """
    if not isinstance(map_in, maps.FlatSkyMap):
        raise TypeError("Input must be a FlatSkyMap")

    # Construct the output map
    if map_stub is None:
        kwargs.setdefault("coord_ref", map_in.coord_ref)
        kwargs.setdefault("pol_type", map_in.pol_type)
        kwargs.setdefault("pol_conv", map_in.pol_conv)
        map_out = maps.HealpixSkyMap(**kwargs)
    else:
        if not isinstance(map_stub, maps.HealpixSkyMap):
            raise TypeError("Output stub must be a HealpixSkyMap")
        map_out = map_stub.clone(False)

    # optimize ringsparse storage
    a0 = map_in.alpha_center
    da = map_in.x_res * map_in.shape[1]
    circ = 2 * np.pi * core.G3Units.rad
    a0 = np.mod(a0 + circ, circ)
    amin = np.mod(a0 - da + circ, circ)
    amax = np.mod(a0 + da + circ, circ)
    amin_shift = np.mod(a0 - da + circ + circ / 2, circ)
    amax_shift = np.mod(a0 + da + circ + circ / 2, circ)
    map_out.shift_ra = bool(np.abs(amax - amin) > np.abs(amax_shift - amin_shift))

    # Populate output map pixels with interpolation and rebinning
    maps.reproj_map(map_in, map_out, rebin=rebin, interp=interp)

    if fullsky:
        map_out.dense = True
        return np.asarray(map_out)
    return map_out
